#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-07-24 18:29:48
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import matplotlib.pyplot as plt


class LossLog():

    def __init__(self, plotdir=None, xlabel='Epoch', ylabel='Loss', title=None, filename=None, logdict=None, lom='min'):
        self.plotdir = plotdir
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.filename = filename
        self.lom = lom
        if logdict is None:
            self.losses = {'train': [], 'valid': [], 'test': []}
        else:
            self.losses = logdict

        self.bests = {}
        for k, v in self.losses.items():
            if len(v) < 1:
                self.bests[k] = float('inf') if self.lom in ['min', 'MIN'] else float('-inf')
            else:
                self.bests[k] = eval(self.lom)(v)

    def assign(self, key, value):
        self.losses[key] = value

    def add(self, key, value):
        self.losses[key].append(value)

    def get(self, key=None):
        return self.losses[key]

    def updir(self, plotdir=None):
        self.plotdir = plotdir

    def plot(self, x=None):
        legend = []
        try:
            plt.figure()
        except Exception as e:
            print(e.args)
            print(str(e))
            print(repr(e))
            import matplotlib; matplotlib.use('Agg');
            import matplotlib.pyplot as plt
            plt.figure()

        for k, v in self.losses.items():
            if len(v) > 0:
                if x is None:
                    plt.plot(v)
                else:
                    plt.plot(x, v)
                legend.append(k)
        plt.legend(legend)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.grid()

        if self.title is not None:
            plt.title(self.title)

        if self.plotdir is None:
            plt.show()
        else:
            if self.filename is None:
                plt.savefig(self.plotdir + '/' + self.ylabel + '_' + self.xlabel + '.png')
            else:
                plt.savefig(self.plotdir + '/' + self.filename)
            plt.close()

    def judge(self, key, n1=50, n2=10):

        loss = self.losses[key]
        n = len(loss)
        flag, proof = False, ''

        if self.lom in ['min', 'MIN']:
            if loss[-1] < self.bests[key]:
                self.bests[key] = loss[-1]
                flag = True
                proof += 'Single'
            if n > n1 + n2:
                if (sum(loss[-n2:]) / n2 <= sum(loss[-n2 - n1:-n2]) / n1) and (loss[-n2:].index(min(loss[-n2:])) == n2 - 1):
                    flag = True
                    proof += 'Average'

        return flag, proof


if __name__ == '__main__':

    loslog = LossLog(plotdir='./', xlabel='xlabel', ylabel='ylabel')
    loslog = LossLog(plotdir=None, xlabel='Epoch', ylabel='Loss', title=None, filename='LossEpoch', logdict={'train': [], 'valid': []})
    for n in range(100):
        loslog.add('train', n)
        loslog.add('valid', n - 1)

    loslog.plot()



from __future__ import division, print_function, absolute_import
from .version import __version__


from .base import baseops, arrayops, mathops, randomfunc
from .base.baseops import dreplace, dmka
from .base.arrayops import sl, cut, cat, arraycomb
from .base.mathops import nextpow2, prevpow2, ebeo, r2c, c2r, conj, real, imag, abs, pow
from .base.randomfunc import setseed, randgrid, randperm, randperm2d


from .utils.const import *
from .utils.colors import rgb2gray, gray2rgb, DISTINCT_COLORS_HEX, DISTINCT_COLORS_RGB, DISTINCT_COLORS_CMYK, DISTINCT_COLORS_RGB_NORM, BASE_COLORS, TABLEAU_COLORS, CSS4_COLORS
from .utils.colormaps import cmaps, viridis, parula
from .utils.convert import str2list, str2num, str2sec
from .utils.ios import loadyaml, loadjson, loadmat, savemat, loadh5, saveh5, mvkeyh5
from .utils.image import imread, imsave, histeq, imresize
from .utils.file import listxfile, pathjoin, fileparts, readtxt, readnum, readcsv, readsec
from .utils.plot_show import cplot, plots, Plots
from .utils.typevalue import bin2int, peakvalue

from .summary.loss_log import LossLog

from .evaluation.contrast import contrast
from .evaluation.entropy import entropy
from .evaluation.norm import fnorm, pnorm
from .evaluation.error import mse, sse, mae, sae
from .evaluation.snr import snr, psnr
from .evaluation.detection_voc import bbox_iou, calc_detection_voc_ap, calc_detection_voc_prec_rec, eval_detection_voc

from .compression.huffman_coding import HuffmanCoding

from .dsp.ffts import padfft, freq, fftfreq, fftshift, ifftshift, fft, ifft, fftx, ffty, ifftx, iffty
from .dsp.convolution import conv1, cutfftconv1, fftconv1
from .dsp.correlation import corr1, cutfftcorr1, fftcorr1
from .dsp.normalsignals import rect, chirp
from .dsp.interpolation1d import sinc, sinc_table, sinc_interp, interp
from .dsp.interpolation2d import interp2d
from .dsp.function_base import unwrap, unwrap2

from .misc.transform import standardization, scale, quantization, ct2rt, rt2ct, db20
from .misc.mapping_operation import mapping
from .misc.sampling import slidegrid, dnsampling, sample_tensor, shuffle_tensor, split_tensor, tensor2patch, patch2tensor, read_samples
from .misc.bounding_box import plot_bbox, fmt_bbox
from .misc.draw_shapes import draw_rectangle



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2020-03-23 19:28:33
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import
import numpy as np
from pyailib.base.arrayops import sl, cat
from pyailib.base.randomfunc import setseed, randgrid, randperm
from pyailib.base.arrayops import arraycomb
from pyailib.utils.ios import loadmat, loadh5


def upsampling(X, shape, axis=-1, method='Lanczos'):

    # Na, Nr = X.shape
    # Y = np.zeros(shape, dtype=X.dtype)
    # for a in range(Na):
    #     Y[a, :] = np.interp(range(shape[1]), range(Nr), X[a, :])
    # for r in range(Nr):
    #     Y[:, a] = np.interp(range(shape[0]), range(Na), X[:, r])

    # return Y

    print(shape, X.shape)
    imgXr = Image.fromarray(X[:, :, 0])
    # imgXr = imgXr.resize((shape[1], shape[0]), Image.LANCZOS)
    # imgXr = imgXr.resize((shape[1], shape[0]), Image.BILINEAR)
    imgXr = imgXr.resize((shape[1], shape[0]), Image.NEAREST)
    # imgXr = imgXr.resize((shape[1], shape[0]), Image.ANTIALIAS)
    imgXi = Image.fromarray(X[:, :, 1])
    # imgXi = imgXi.resize((shape[1], shape[0]), Image.LANCZOS)
    # imgXi = imgXi.resize((shape[1], shape[0]), Image.BILINEAR)
    imgXi = imgXi.resize((shape[1], shape[0]), Image.NEAREST)
    # imgXi = imgXi.resize((shape[1], shape[0]), Image.ANTIALIAS)
    return np.transpose(np.array([np.array(imgXr), np.array(imgXi)]), (1, 2, 0))


def slidegrid(start, stop, step, shake=0, n=None):
    r"""generates sliding grid indexes

    Generates :attr:`n` sliding grid indexes from :attr:`start` to :attr:`stop`
    with step size :attr:`step`.

    Args:
        start (int or list): start sampling point
        stop (int or list): stop sampling point
        step (int or list): sampling stepsize
        shake (float): the shake rate, if :attr:`shake` is 0, no shake, (default),
            if positive, add a positive shake, if negative, add a negative.
        n (int or None): the number of samples (default None, int((stop0 - start0) / step0) * int((stop1 - start1) / step1)...).

    Returns:
        for multi-dimension, return a 2-d tensor, for 1-dimension, return a 1d-tensor.

    Raises:
        TypeError: The number of samples should be an integer or None.

    see :func:`randperm`, :func:`randgrid`.

    """

    starts = [start] if type(start) is int else start
    stops = [stop] if type(stop) is int else stop
    steps = [step] if type(step) is int else step
    shakes = [shake] if type(shake) is int or type(shake) is float else shake
    if (n is not None) and (type(n) is not int):
        raise TypeError('The number of samples should be an integer or None!')
    elif n is None:
        n = float('inf')
    index = []
    for start, stop, step, shake in zip(starts, stops, steps, shakes):
        shakep = shake if abs(shake) >= 1 and type(shake) is int else int(shake * step)
        x = np.array(range(start, stop, step))
        if shakep != 0:
            s = np.random.randint(0, abs(shakep), len(x))
            x = x - s if shakep < 0 else x + s
            x[x >= (stop - step)] = stop - step
            x[x < start] = start
        index.append(x)
    P = arraycomb(index)
    n = min(P.shape[0], n)
    P = P[:n, ...]

    if len(starts) == 1:
        P = P.squeeze(1)
        return P
    else:
        return P.transpose()


def dnsampling(x, ratio=1., axis=-1, smode='uniform', omode='discard', seed=None, extra=False):
    """down-sampling a tensor

    Args:
        x (ndarray): The input tensor.
        ratio (float, optional): Downsampling ratio.
        axis (int, optional): Downsampling axis (default -1).
        smode (str, optional): Downsampling mode: ``'uniform'``, ``'random'``, ``'random2'``.
        omode (str, optional): output mode: ``'discard'`` for discarding, ``'zero'`` for zero filling.
        seed (int or None, optional): seed for numpy's random.
        extra (bool, optional): If ``True``, also return sampling mask.

    Returns:
        (ndarray): Description

    Raises:
        TypeError: :attr:`axis`
        ValueError: :attr:`ratio`, attr:`smode`, attr:`omode`
    """

    nDims = np.ndim(x)
    if type(axis) is int:
        if type(ratio) is not float:
            raise ValueError('Downsampling ratio should be a number!')
        axis = [axis]
        ratio = [ratio]
    elif type(axis) is list or tuple:
        if len(axis) != len(ratio):
            raise ValueError('You should specify the DS ratio for each axis!')
    else:
        raise TypeError('Wrong type of axis!')

    axis, ratio = list(axis), list(ratio)
    for cnt in range(len(axis)):
        if axis[cnt] < 0:
            axis[cnt] += nDims
        # ratio[cnt] = 1. - ratio[cnt]
        cnt += 1

    if omode in ['discard', 'DISCARD', 'Discard']:
        if smode not in ['uniform', 'UNIFORM', 'Uniform']:
            raise ValueError("Only support uniform mode!")

        index = [slice(None)] * nDims
        for a, r in zip(axis, ratio):
            sa = x.shape[a]
            da = int(round(1. / r))
            index[a] = slice(0, sa, da)
        index = tuple(index)

        if extra:
            return x[index], index
        else:
            return x[index]

    elif omode in ['zero', 'ZERO', 'Zeros']:
        mshape = [1] * nDims
        for a in axis:
            mshape[a] = x.shape[a]
        mask = np.zeros(mshape, dtype=np.uint8)
        if smode in ['uniform', 'UNIFORM', 'Uniform']:
            for a, r in zip(axis, ratio):
                sa = x.shape[a]
                da = int(round(1. / r))
                idx = sl(nDims, a, slice(0, sa, da))
                mask[idx] += 1
            mask[mask < len(axis)] = 0
            mask[mask >= len(axis)] = 1

        elif smode in ['random', 'RANDOM', 'Random']:
            setseed(seed, target='numpy')
            for a, r in zip(axis, ratio):
                d = np.ndim(x)
                s = x.shape[a]
                n = int(round(s * r))
                idx = randperm(0, s, n)
                idx = np.sort(idx)
                idx = sl(d, a, idx)
                mask[idx] += 1
            mask[mask < len(axis)] = 0
            mask[mask >= len(axis)] = 1

        elif smode in ['random2', 'RANDOM2', 'Random2']:
            setseed(seed, target='numpy')
            d = np.ndim(x)
            s0, s1 = x.shape[axis[0]], x.shape[axis[1]]
            n0, n1 = int(round(s0 * ratio[0])), int(round(s1 * ratio[0]))
            idx0 = randperm(0, s0, n0)
            # idx0 = np.sort(idx0)

            for i0 in idx0:
                idx1 = randperm(0, s1, n1)
                mask[sl(d, [axis[0], axis[1]], [[i0], idx1])] = 1

        else:
            raise ValueError('Not supported sampling mode: %s!' % smode)

        if extra:
            return x * mask, mask
        else:
            return x * mask

    else:
        raise ValueError('Not supported output mode: %s!' % omode)


def sample_tensor(x, n, axis=0, groups=1, mode='sequentially', seed=None, extra=False):
    """sample a tensor

    Sample a tensor sequentially/uniformly/randomly.

    Args:
        x (ndarray): a numpy or torch tensor to be sampled
        n (int): sample number
        axis (int, optional): the axis to be sampled (the default is 0)
        groups (int, optional): number of groups in this tensor (the default is 1)
        mode (str, optional): - ``'sequentially'``: evenly spaced (default)
            - ``'uniformly'``: [0, int(n/groups)]
            - ``'randomly'``: randomly selected, non-returned sampling
        seed (None or int, optional): only work for ``'randomly'`` mode (the default is None)
        extra (bool, optional): If ``True``, also return the selected indexes, the default is ``False``.

    Returns:
        y (ndarray): Sampled numpy or torch tensor.
        idx (list): Sampled indexes, if :attr:`extra` is ``True``, this will also be returned.


    Example:

        ::

            setseed(2020, 'numpy')

            x = np.randint(1000, (20, 3, 4))
            y1, idx1 = sample_tensor(x, 10, axis=0, groups=2, mode='sequentially', extra=True)
            y2, idx2 = sample_tensor(x, 10, axis=0, groups=2, mode='uniformly', extra=True)
            y3, idx3 = sample_tensor(x, 10, axis=0, groups=2, mode='randomly', extra=True)

            print(x.shape)
            print(y1.shape)
            print(y2.shape)
            print(y3.shape)
            print(idx1)
            print(idx2)
            print(idx3)

            the outputs are as follows:

            torch.Size([20, 3, 4])
            torch.Size([10, 3, 4])
            torch.Size([10, 3, 4])
            torch.Size([10, 3, 4])
            [0, 1, 2, 3, 4, 10, 11, 12, 13, 14]
            [0, 2, 4, 6, 8, 10, 12, 14, 16, 18]
            [3, 1, 5, 8, 7, 17, 18, 13, 16, 10]


    Raises:
        ValueError: The tensor does not has enough samples.


    """

    N = x.shape[axis]
    M = int(N / groups)  # each group has M samples
    m = int(n / groups)  # each group has m sampled samples

    if (M < m):
        raise ValueError('The ndarray does not has enough samples')

    idx = []
    if mode in ['sequentially', 'Sequentially']:
        for g in range(groups):
            idx += list(range(int(M * g), int(M * g) + m))
    if mode in ['uniformly', 'Uniformly']:
        for g in range(groups):
            idx += list(range(int(M * g), int(M * g + M), int(M / m)))[:m]
    if mode in ['randomly', 'Randomly']:
        setseed(seed, target='numpy')
        for g in range(groups):
            idx += list(randperm(int(M * g), int(M * g + M), m))

    if extra:
        return x[sl(np.ndim(x), axis=axis, idx=[idx])], idx
    else:
        return x[sl(np.ndim(x), axis=axis, idx=[idx])]


def shuffle_tensor(x, axis=0, groups=1, mode='inter', seed=None, extra=False):
    """shuffle a tensor

    Shuffle a tensor randomly.

    Args:
        x (ndarray): A numpy or torch tensor to be shuffled.
        axis (int, optional): The axis to be shuffled (default 0)
        groups (number, optional): The number of groups in this tensor (default 1)
        mode (str, optional):
            - ``'inter'``: between groups (default)
            - ``'intra'``: within group
            - ``'whole'``: the whole
        seed (None or number, optional): random seed (the default is None)
        extra (bool, optional): If ``True``, also returns the shuffle indexes, the default is ``False``.

    Returns:
        y (ndarray): Shuffled numpy or torch tensor.
        idx (list): Shuffled indexes, if :attr:`extra` is ``True``, this will also be returned.


    Example:

        Shuffle a tensor randomly with different modes (``'intra'``, ``'inter'``, ``'whole'``).

        ::

            setseed(2020, 'numpy')

            x = np.randint(1000, (20, 3, 4))
            y1, idx1 = shuffle_tensor(x, axis=0, groups=4, mode='intra', extra=True)
            y2, idx2 = shuffle_tensor(x, axis=0, groups=4, mode='inter', extra=True)
            y3, idx3 = shuffle_tensor(x, axis=0, groups=4, mode='whole', extra=True)

            print(x.shape)
            print(y1.shape)
            print(y2.shape)
            print(y3.shape)
            print(idx1)
            print(idx2)
            print(idx3)

            the outputs are as follows:

            torch.Size([20, 3, 4])
            torch.Size([20, 3, 4])
            torch.Size([20, 3, 4])
            torch.Size([20, 3, 4])
            [1, 0, 3, 4, 2, 8, 6, 5, 9, 7, 13, 11, 12, 14, 10, 18, 15, 17, 16, 19]
            [0, 1, 2, 3, 4, 10, 11, 12, 13, 14, 5, 6, 7, 8, 9, 15, 16, 17, 18, 19]
            [1, 13, 12, 5, 19, 9, 11, 6, 4, 16, 17, 3, 8, 18, 7, 10, 15, 0, 14, 2]


    """

    N = x.shape[axis]
    M = int(N / groups)  # each group has M samples

    idx = []
    setseed(seed, target='numpy')
    if mode in ['whole', 'Whole', 'WHOLE']:
        idx = list(randperm(0, N, N))

    if mode in ['intra', 'Intra', 'INTRA']:
        for g in range(groups):
            idx += list(randperm(int(M * g), int(M * g + M), M))
    if mode in ['inter', 'Inter', 'INTER']:
        for g in range(groups):
            idx += [list(range(int(M * g), int(M * g + M)))]
        groupidx = list(randperm(0, groups, groups))

        iidx = idx.copy()
        idx = []
        for i in groupidx:
            idx += iidx[i]

    if extra:
        return x[sl(np.ndim(x), axis=axis, idx=[idx])], idx
    else:
        return x[sl(np.ndim(x), axis=axis, idx=[idx])]


def split_tensor(x, ratios=[0.7, 0.2, 0.1], axis=0, shuffle=False, seed=None, extra=False):
    """split a tensor

    split a tensor into some parts.

    Args:
        x (ndarray): A numpy array or torch tensor.
        ratios (list, optional): Split ratios (the default is [0.7, 0.2, 0.05])
        axis (int, optional): Split axis (the default is 0)
        shuffle (bool, optional): Whether shuffle (the default is False)
        seed (int, optional): Shuffule seed (the default is None)
        extra (bool, optional): If ``True``, also return the split indexes, the default is ``False``.

    Returns:
        (list of ndarray): Splitted ndarrays.
    """

    y, idxes = [], []

    N, ns = x.shape[axis], 0
    if shuffle:
        setseed(seed, target='numpy')
        idx = randperm(0, N, N)
    else:
        idx = list(range(N))

    for ratio in ratios:
        n = int(ratio * N)
        idxes.append(idx[ns:ns + n])
        y.append(x[idx[ns:ns + n]])
        ns += n

    if extra:
        return y, idxes
    else:
        return y


def tensor2patch(x, n=None, size=(256, 256), axis=(0, 1), start=(0, 0), stop=(None, None), step=(1, 1), shake=(0, 0), mode='slidegrid', seed=None):
    """sample patch from a tensor

    Sample some patches from a tensor, tensor and patch can be any size.

    Args:
        x (ndarray): A tensor to be sampled.
        n (int, optional): The number of pactches, the default is None, auto computed,
            equals to the number of blocks with specified :attr:`step`
        size (tuple or int, optional): The size of patch (the default is (256, 256))
        axis (tuple or int, optional): The sampling axis (the default is (0, 1))
        start (tuple or int, optional): Start sampling index for each axis (the default is (0, 0))
        stop (tuple or int, optional): Stopp sampling index for each axis. (the default is (None, None), which [default_description])
        step (tuple or int, optional): Sampling stepsize for each axis  (the default is (1, 1), which [default_description])
        shake (tuple or int or float, optional): float for shake rate, int for shake points (the default is (0, 0), which means no shake)
        mode (str, optional): Sampling mode, ``'slidegrid'``, ``'randgrid'``, ``'randperm'`` (the default is 'slidegrid')
        seed (int, optional): Random seed. (the default is None, which means no seed.)

    Returns:
        (ndarray): A tensor of sampled patches.
    
    see :func:`patch2tensor`.

    Example:

        Sample patches from a tensor with different mode (randperm, randgrid, slidegrid), 
        and then reform these patches into an image.

        .. image:: ./_static/demo_sample_patch.png
        :scale: 100 %
        :align: center

        The results shown in the above figure can be obtained by the following codes.

        ::

            import math
            import numpy as np
            import pyailib as pl
            import matplotlib.pyplot as plt

            filename = '../../data/images/Lotus512.png'
            filename = '../../data/images/LenaRGB512.tif'

            x = pl.imread(filename)
            xshape = x.shape
            xshape = xshape[:2]

            n, size = 64, (64, 64)

            y1 = pl.tensor2patch(x, n=n, size=size, axis=(0, 1), step=(1, 1), shake=(0, 0), mode='randperm', seed=2020)
            y2 = pl.tensor2patch(x, n=n, size=size, axis=(0, 1), step=(64, 64), shake=(0, 0), mode='randgrid', seed=2020)
            y3 = pl.tensor2patch(x, n=n, size=size, axis=(0, 1), step=(64, 64), shake=(0, 0), mode='slidegrid', seed=2020)
            y4 = pl.tensor2patch(x, n=n, size=size, axis=(0, 1), step=(64, 64), shake=(32, 32), mode='slidegrid', seed=2020)

            print(y1.shape, y2.shape, y3.shape, y4.shape)

            Y1 = pl.patch2tensor(y1, size=xshape, axis=(1, 2), mode='nfirst')
            Y2 = pl.patch2tensor(y2, size=xshape, axis=(1, 2), mode='nfirst')
            Y3 = pl.patch2tensor(y3, size=xshape, axis=(1, 2), mode='nfirst')
            Y4 = pl.patch2tensor(y4, size=xshape, axis=(1, 2), mode='nfirst')

            plt.figure()
            plt.subplot(221)
            plt.imshow(Y1)
            plt.title('randperm, shake=(0, 0)')
            plt.subplot(222)
            plt.imshow(Y2)
            plt.title('randgrid, shake=(0, 0)')
            plt.subplot(223)
            plt.imshow(Y3)
            plt.title('slidegrid, shake=(0, 0)')
            plt.subplot(224)
            plt.imshow(Y4)
            plt.title('slidegrid, shake=(32, 32)')
            plt.show()

    """

    axis = [axis] if type(axis) is int else list(axis)
    naxis = len(axis)
    sizep = [size] * naxis if type(size) is int else list(size)
    start = [start] * naxis if type(start) is int else list(start)
    stop = [stop] * naxis if type(stop) is int else list(stop)
    step = [step] * naxis if type(step) is int else list(step)
    shake = [shake] * naxis if type(shake) is float else list(shake)

    dimx = np.ndim(x)
    dimp = len(axis)
    sizex = np.array(x.shape)
    sizep = np.array(sizep)

    npatch = []
    npatch = np.uint32(sizex[axis] / sizep)
    N = int(np.prod(npatch))
    n = N if n is None else int(n)

    yshape = list(x.shape)
    for a, p in zip(axis, sizep):
        yshape[a] = p
    yshape = [n] + yshape

    for i in range(naxis):
        if stop[i] is None:
            stop[i] = sizex[axis[i]]

    y = np.zeros(yshape, dtype=x.dtype)

    if mode in ['slidegrid', 'SLIDEGRID', 'SlideGrid']:
        assert n <= N, ('n should be slower than ' + str(N + 1))
        seppos = slidegrid(start, stop, step, shake, n)
    if mode in ['randgrid', 'RANDGRID', 'RandGrid']:
        assert n <= N, ('n should be slower than ' + str(N + 1))
        setseed(seed, target='numpy')
        seppos = randgrid(start, stop, step, shake, n)

    if mode in ['randperm', 'RANDPERM', 'RandPerm']:
        setseed(seed, target='numpy')
        stop = [x - y for x, y in zip(stop, sizep)]
        seppos = randgrid(start, stop, [1] * dimp, [0] * dimp, n)

    for i in range(n):
        indexi = []
        for j in range(dimp):
            indexi.append(slice(seppos[j][i], seppos[j][i] + sizep[j]))
        t = x[sl(dimx, axis, indexi)]
        y[i] = t
    return y


def patch2tensor(p, size=(256, 256), axis=(1, 2), mode='nfirst'):
    """merge patch to a tensor


    Args:
        p (tensor): A tensor of patches.
        size (tuple, optional): Merged tensor size in the dimension (the default is (256, 256)).
        axis (tuple, optional): Merged axis of patch (the default is (1, 2))
        mode (str, optional): Patch mode ``'nfirst'`` or ``'nlast'`` (the default is 'nfirst',
            which means the first dimension is the number of patches)

    Returns:
        ndarray: Merged tensor.
    
    see :func:`tensor2patch`.

    """

    naxis = len(axis)
    sizep = list(p.shape)
    sizex = list(p.shape)
    dimp = np.ndim(p)
    axisp = list(range(0, dimp))
    npatch = []
    steps = sizep.copy()
    for a, s in zip(axis, size):
        npatch.append(int((s * 1.) / sizep[a]))
        sizex[a] = s
        steps[a] = sizep[a]

    if mode in ['nfirst', 'Nfirst', 'NFIRST']:
        axisn = 0
        N = p.shape[0]
        sizex = sizex[1:]
        steps = sizep[1:]
    if mode in ['nlast', 'Nlast', 'NLAST']:
        axisn = -1
        N = p.shape[-1]
        sizex = sizex[:-1]
        steps = sizep[:-1]

    x = np.zeros(sizex, dtype=p.dtype)

    dimx = np.ndim(x)
    axisx = list(range(dimx))

    index = []
    for a, stop, step in zip(axisx, sizex, steps):
        idx = np.array(range(0, stop, step))
        index.append(idx)
    index = arraycomb(index)
    naxisx = len(axisx)
    for n in range(N):
        indexn = []
        for a in axisx:
            indexn.append(slice(index[n, a], index[n, a] + steps[a], 1))
        x[sl(dimx, axisx, indexn)] = p[sl(dimp, axisn, n)]
    return x


def read_samples(datafiles, keys=[['SI', 'ca', 'cr']], nsamples=[10], groups=[1], mode='sequentially', axis=0, parts=None, seed=None):
    """Read samples

    Args:
        datafiles (list): list of path strings
        keys (list, optional): data keys to be read
        nsamples (list, optional): number of samples for each data file
        groups (list, optional): number of groups in each data file
        mode (str, optional): sampling mode for all datafiles
        axis (int, optional): sampling axis for all datafiles
        parts (None, optional): number of parts (split samples into some parts)
        seed (None, optional): the seed for random stream

    Returns:
        tensor: samples

    Raises:
        ValueError: :attr:`nsamples` should be large enough
    """

    nfiles = len(datafiles)
    if len(keys) == 1:
        keys = keys * nfiles
    if len(nsamples) == 1:
        nsamples = nsamples * nfiles
    if len(groups) == 1:
        groups = groups * nfiles

    nkeys = len(keys[0])

    if parts is None:
        outs = [[]] * nkeys
    else:
        nparts = len(parts)
        outs = [[[]] * nparts] * nkeys

    for datafile, key, n, group in zip(datafiles, keys, nsamples, groups):

        if datafile[datafile.rfind('.'):] == '.mat':
            data = loadmat(datafile)
        if datafile[datafile.rfind('.'):] in ['.h5', '.hdf5']:
            data = loadh5(datafile)

        N = data[key[0]].shape[axis]
        M = int(N / group)  # each group has M samples
        m = int(n / group)  # each group has m sampled samples

        if (M < m):
            raise ValueError('The tensor does not has enough samples')

        idx = []
        if mode in ['sequentially', 'Sequentially']:
            for g in range(group):
                idx += list(range(int(M * g), int(M * g) + m))
        if mode in ['uniformly', 'Uniformly']:
            for g in range(group):
                idx += list(range(int(M * g), int(M * g + M), int(M / m)))[:m]
        if mode in ['randomly', 'Randomly']:
            setseed(seed)
            for g in range(group):
                idx += randperm(int(M * g), int(M * g + M), m)

        for j, k in enumerate(key):
            d = np.ndim(data[k])
            if parts is None:
                outs[j] = cat((outs[j], data[k][sl(d, axis, [idx])]), axis=axis)
            else:
                nps, npe = 0, 0
                for i in range(nparts):
                    part = parts[i]
                    npe = nps + int(part * group)
                    outs[j][i] = cat((outs[j][i], data[k][sl(d, axis, [idx[nps:npe]])]), axis=axis)
                    nps = npe

    return outs


if __name__ == '__main__':

    setseed(2020)
    x = np.random.randint(0, 1000, (20, 3, 4))

    y1, idx1 = sample_tensor(x, 10, axis=0, groups=2, mode='sequentially', extra=True)
    y2, idx2 = sample_tensor(x, 10, axis=0, groups=2, mode='uniformly', extra=True)
    y3, idx3 = sample_tensor(x, 10, axis=0, groups=2, mode='randomly', extra=True)

    print(x.shape)
    print(y1.shape)
    print(y2.shape)
    print(y3.shape)
    print(idx1)
    print(idx2)
    print(idx3)

    x = np.random.randint(0, 1000, (20, 3, 4))
    y1, idx1 = shuffle_tensor(x, axis=0, groups=4, mode='intra', extra=True)
    y2, idx2 = shuffle_tensor(x, axis=0, groups=4, mode='inter', extra=True)
    y3, idx3 = shuffle_tensor(x, axis=0, groups=4, mode='whole', extra=True)

    print(x.shape)
    print(y1.shape)
    print(y2.shape)
    print(y3.shape)
    print(idx1)
    print(idx2)
    print(idx3)

    y1, y2, y3 = split_tensor(x, ratios=[0.7, 0.2, 0.1], axis=0, shuffle=False, seed=None)
    print(y3)
    y1, y2, y3 = split_tensor(x, ratios=[0.7, 0.2, 0.1], axis=0, shuffle=True, seed=None)
    print(y3)
    y1, y2, y3 = split_tensor(x, ratios=[0.7, 0.2, 0.1], axis=0, shuffle=True, seed=2021)
    print(y3)
    y1, y2, y3 = split_tensor(x, ratios=[0.7, 0.2, 0.1], axis=0, shuffle=True, seed=2021)
    print(y3)
    print(y1.shape, y2.shape, y3.shape)

    Na, Nr, Nc = (9, 12, 2)
    x = np.random.randint(0, 1000, (Na, Nr, Nc))

    print(x[:, :, 0], 'x', x.shape)
    print(x[:, :, 1], 'x', x.shape)

    y = dnsampling(x, ratio=(0.5, 0.5), axis=(0, 1), smode='uniform', omode='discard')
    print(y[:, :, 0], 'discard')
    print(y[:, :, 1], 'discard')

    y = dnsampling(x, ratio=(0.5, 0.5), axis=(0, 1), smode='uniform', omode='zero')
    print(y[:, :, 0], 'zero')
    print(y[:, :, 1], 'zero')

    y = dnsampling(x, ratio=(0.5, 0.5), axis=(0, 1), smode='random', omode='zero')
    print(y[:, :, 0], 'zero')
    print(y[:, :, 1], 'zero')

    y = dnsampling(x, ratio=(0.5, 0.5), axis=(0, 1), smode='random2', omode='zero')
    print(y[:, :, 0], 'zero')
    print(y[:, :, 1], 'zero')

    # y = tensor2patch(x, n=None, size=(2, 3), axis=(0, 1), mode='slide', step=(1, 1), seed=None)
    # print(y.shape, 'slide')
    # print(y[0, :, :, 0], 'slide')
    # print(y[0, :, :, 1], 'slide')
    y = tensor2patch(x, n=None, size=(2, 3), axis=(0, 1), step=(2, 3), shake=(0, 0), mode='randgrid', seed=None)
    print(y.shape, 'randgrid')
    print(y[0, :, :, 0], 'randgrid')
    print(y[0, :, :, 1], 'randgrid')

    y = tensor2patch(x, n=None, size=(2, 3), axis=(0, 1), step=(1, 1), shake=(0, 0), mode='randperm', seed=None)
    print(y.shape, 'randperm')
    print(y[0, :, :, 0], 'randperm')
    print(y[0, :, :, 1], 'randperm')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-02-25 09:53:21
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

from __future__ import division, absolute_import


import numpy as np
from pyailib.evaluation.error import mse
from pyailib.utils.typevalue import peakvalue


def snr():
    pass


def psnr(P, G, caxis=None, axis=None, vpeak=None, reduction=None):
    r"""Peak Signal-to-Noise Ratio

    The Peak Signal-to-Noise Ratio (PSNR) is expressed as

    .. math::
        {\rm PSNR} = 10 \log10(\frac{V_{\rm peak}^2}{\rm MSE})

    For float data, :math:`V_{\rm peak} = 1`;

    For interges, :math:`V_{\rm peak} = 2^{\rm nbits}`,
    e.g. uint8: 255, uint16: 65535 ...


    Parameters
    -----------
    P : array_like
        The data to be compared. For image, it's the reconstructed image.
    G : array_like
        Reference data array. For image, it's the original image.
    vpeak : float, int or None, optional
        The peak value. If None, computes automaticly.
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is :obj:`None`)

    Returns
    -------
    PSNR : float
        Peak Signal to Noise Ratio value.

    Examples
    ---------

    ::

        P = np.array([[0, 200, 210], [220, 5, 6]])
        G = np.array([[251, 200, 210], [220, 5, 6]])
        PSNR = psnr(P, G, vpeak=None)
        print(PSNR)

        P = np.array([[251, 200, 210], [220, 5, 6]]).astype('uint8')
        G = np.array([[0, 200, 210], [220, 5, 6]]).astype('uint8')
        PSNR = psnr(P, G, vpeak=None)
        print(PSNR)

        P = np.array([[251, 200, 210], [220, 5, 6]]).astype('float')
        G = np.array([[0, 200, 210], [220, 5, 6]]).astype('float')
        PSNR = psnr(P, G, vpeak=None)
        print(PSNR)

    """

    if P.dtype != G.dtype:
        print("Warning: P(" + str(P.dtype) + ")and G(" + str(G.dtype) +
              ")have different type! PSNR may not right!")

    if vpeak is None:
        vpeak = peakvalue(G)

    MSE = mse(P, G, caxis=caxis, axis=axis, reduction=reduction)
    PSNR = 10 * np.log10((vpeak ** 2) / MSE)

    return PSNR


if __name__ == '__main__':

    P = np.array([[0, 200, 210], [220, 5, 6]])
    G = np.array([[251, 200, 210], [220, 5, 6]])
    PSNR = psnr(P, G, vpeak=None)
    print(PSNR)

    P = np.array([[251, 200, 210], [220, 5, 6]]).astype('uint8')
    G = np.array([[0, 200, 210], [220, 5, 6]]).astype('uint8')
    PSNR = psnr(P, G, vpeak=None)
    print(PSNR)

    P = np.array([[251, 200, 210], [220, 5, 6]]).astype('float')
    G = np.array([[0, 200, 210], [220, 5, 6]]).astype('float')
    PSNR = psnr(P, G, vpeak=None)
    print(PSNR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2020-03-23 19:28:33
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import
import numpy as np


def sl(dims, axis, idx=None):
    r"""slice any axis

    generates slice of specified axis.

    Parameters
    ----------
    dims : int
        total dimensions
    axis : int or list
        select axis list.
    idx : list or None, optional
        slice lists of the specified :attr:`axis`, if None, does nothing (the default)

    Returns
    -------
    tuple of slice
        slice for specified axis elements.

    Examples
    --------

    ::

        import numpy as np

        np.random.seed(2020)
        X = np.random.randint(0, 100, (9, 10))
        print(X, 'X)
        print(X[sl(2, -1, [0, 1])], 'Xsl')

        # output:

        [[96  8 67 67 91  3 71 56 29 48]
        [32 24 74  9 51 11 55 62 67 69]
        [48 28 20  8 38 84 65  1 79 69]
        [74 73 62 21 29 90  6 38 22 63]
        [21 68  6 98  3 20 55  1 52  9]
        [83 82 65 42 66 55 33 80 82 72]
        [94 91 14 14 75  5 38 83 99 10]
        [80 64 79 30 84 22 46 26 60 13]
        [24 63 25 89  9 69 47 89 55 75]] X
        [[96  8]
        [32 24]
        [48 28]
        [74 73]
        [21 68]
        [83 82]
        [94 91]
        [80 64]
        [24 63]] Xsl

    """

    idxall = [slice(None)] * dims

    axis = [axis] if type(axis) is int else axis
    idx = [idx] if type(idx) not in [list, tuple] else idx
    if len(axis) != len(idx):
        raise ValueError('The index for each axis should be given!')

    naxis = len(axis)
    for n in range(naxis):
        idxall[axis[n]] = idx[n]

    return tuple(idxall)


def cut(x, pos, axis=None):
    r"""Cut array at given position.

    Cut array at given position.

    Parameters
    ----------
    x : numpy array
        an array to be cutted.
    pos : tulpe or list
        cut positions: ((cpstart, cpend), (cpstart, cpend), ...)
    axis : int, tulpe or list, optional
        cut axis (the default is None, which means nothing)
    """

    if axis is None:
        return x
    if type(axis) == int:
        axis = tuple([axis])
    nDims = np.ndim(x)
    idx = [None] * nDims

    if len(axis) > 1 and len(pos) != len(axis):
        raise ValueError('You should specify cut axis for each cut axis!')
    elif len(axis) == 1:
        axis = tuple(list(axis) * len(pos))

    uqaixs = np.unique(axis)
    for a in uqaixs:
        idx[a] = []

    for i in range(len(axis)):
        idx[axis[i]] += range(pos[i][0], pos[i][1])

    for a in uqaixs:
        idxall = [slice(None)] * nDims
        idxall[a] = idx[a]
        x = x[tuple(idxall)]
    return x


def cat(arrays, axis=None, out=None):

    arrays = list(arrays)
    for n in range(len(arrays)):
        try:
            arrays.remove([])
            arrays.remove(None)
        except:
            pass
    return np.concatenate(arrays, axis=axis, out=out)


def arraycomb(arrays, out=None):
    r"""compute the elemnts combination of several lists.

    Args:
        arrays (list or numpy array): The lists or arrays.
        out (numpy array, optional): The combination results (defaults is :obj:`None`).

    Returns:
        numpy array: The combination results.

    Examples:

    Compute the combination of three lists: :math:`[1,2,3]`, :math:`[4, 5]`, :math:`[6,7]`,
    this will produce a :math:`12\times 3` array.

    ::

        x = arraycomb(([1, 2, 3], [4, 5], [6, 7]))
        print(x, x.shape)

        # output:
        [[1 4 6]
        [1 4 7]
        [1 5 6]
        [1 5 7]
        [2 4 6]
        [2 4 7]
        [2 5 6]
        [2 5 7]
        [3 4 6]
        [3 4 7]
        [3 5 6]
        [3 5 7]] (12, 3)

    """
    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype
    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)
    m = int(n / arrays[0].size)
    out[:, 0] = np.repeat(arrays[0], m)

    if arrays[1:]:
        arraycomb(arrays[1:], out=out[0:m, 1:])

    for j in range(1, arrays[0].size):
        out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]

    return out


if __name__ == '__main__':

    np.random.seed(2020)
    X = np.random.randint(0, 100, (9, 10))
    print('X')
    print(X)
    Y = cut(X, ((1, 4), (5, 8)), axis=0)
    print('Y = cut(X, ((1, 4), (5, 8)), axis=0)')
    print(Y)
    Y = cut(X, ((1, 4), (7, 9)), axis=(0, 1))
    print('Y = cut(X, ((1, 4), (7, 9)), axis=(0, 1))')
    print(Y)
    Y = cut(X, ((1, 4), (1, 4), (5, 8), (7, 9)), axis=(0, 1, 0, 1))
    print('cut(X, ((1, 4), (1, 4), (5, 8), (7, 9)), axis=(0, 1, 0, 1))')
    print(Y)

    print('--------------------sl')
    print(X, 'X')
    print(X[sl(2, -1, [[0, 1]])], 'Xsl')
    print(X[:, 0:2])
    print('--------------------cat')
    print(cat(([], [], np.array([[1, 2], [3, 4]])), axis=0))

    print('--------------------arraycomb')
    x = arraycomb(([1, 2, 3], [4, 5], [6, 7]))
    print(x, x.shape)

    x = arraycomb(([1, 2, 3, 4], [4, 5], [6, 7]))
    print(x, x.shape)

    x = arraycomb(([1, 2, 3, 4]))
    print(x, x.shape)

    x = arraycomb([[0, 64, 128, 192, 256, 320, 384, 448], [0,  64, 128, 192, 256, 320, 384, 448]])
    print(x, x.shape)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2015-10-15 10:34:16
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.1$

import numpy as np
import codecs


def bin2int(b, endian='<'):
    nb = len(b)
    if endian in ['<', 'Little', 'little']:
        n = int(codecs.encode(b[::-1], 'hex'), 16)
        print(n, nb)
        return n & (2 ** nb - 1)
    if endian in ['>', 'Big', 'big']:
        return int(codecs.encode(b, 'hex'), 16)


def peakvalue(A):
    r"""Compute the peak value of the input.

    Find peak value in matrix

    Parameters
    ----------
    A : numpy array
        Data for finding peak value

    Returns
    -------
    number
        Peak value.
    """

    dtype = A.dtype
    if dtype in ('float', 'float16', 'float32', 'float64', 'float128'):
        Vpeak = 1
    elif dtype in ('uint8', 'uint16', 'uint32', 'uint64'):
        datatype = str(dtype)
        Vpeak = 2 ** int(datatype[4:]) - 1
    elif dtype in ('int64', 'int32', 'int16', 'int8'):
        datatype = str(dtype)
        Vpeak = 2 ** int(datatype[3:]) / 2 - 1
    else:
        logging.info("~~~Unknown type, using the maximum value!")
        Vpeak = np.max(A)

    return Vpeak


if __name__ == '__main__':

    n = -123456
    bs = n.to_bytes(8, 'little', signed=True)
    print(bs)
    print(hex(n))
    print(bin2int(bs, '<'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2020-03-05 16:36:03
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

import numpy as np
from pyailib.dsp.ffts import fft, ifft, padfft
from pyailib.base.mathops import nextpow2
from pyailib.base.arrayops import cut


def corr1(f, g, shape='same'):
    r"""Correlation.

    the correlation between f and g can be expressed as

    .. math::
        (f\star g)[n] = \sum_{m=-\infty}^{+\infty}{\overline{f[m]}g[m+n]} = \sum_{m=-\infty}^{+\infty}\overline{f[m-n]}g[m]
        :label: equ-1DCrossCorrelationDiscrete

    Parameters
    ----------
    f : numpy array
        data1
    g : numpy array
        daat2
    shape : str, optional
        - ``'full'``: returns the full correlation,
        - ``'same'``: returns the central part of the correlation
                that is the same size as f (default).
        - ``'valid'``: returns only those parts of the correlation
                that are computed without the zero-padded edges.
                LENGTH(y)is MAX(LENGTH(f)-MAX(0,LENGTH(g)-1),0).
    """

    Nf = len(f)
    Ng = len(g)
    N = Nf + Ng - 1

    f = np.hstack((np.zeros((Ng - 1)), f, np.zeros((Ng - 1))))
    g = g.conj()
    y = []
    for n in range(N):
        y.append(np.dot(f[n:n + Ng], g))

    if shape in ['same', 'Same', 'SAME']:
        Ns = np.fix(Ng / 2.)
        Ne = Ns + Nf
    if shape in ['valid', 'Valid', 'VALID']:
        Ns = Ng - 1
        Ne = Ns + Nf - Ng + 1
    if shape in ['full', 'Full', 'FULL']:
        Ns = 0
        Ne = N

    Ns, Ne = (int(Ns), int(Ne))

    return np.array(y[Ns:Ne])


def cutfftcorr1(y, nfft, Nx, Nh, shape='same', axis=0, ftshift=False):
    r"""Throwaway boundary elements to get correlation results.

    Throwaway boundary elements to get correlation results.

    Parameters
    ----------
    y : numpy array
        array after ``iff``.
    nfft : int
        number of fft points.
    Nx : int
        signal length
    Nh : int
        filter length
    shape : str
        output shape:
        1. ``'same'`` --> same size as input x, :math:`N_x`
        2. ``'valid'`` --> valid correlation output
        3. ``'full'`` --> full correlation output, :math:`N_x+N_h-1`
        (the default is 'same')
    axis : int
        correlation axis (the default is 0)
    ftshift : bool, optional
        whether to shift the frequencies (the default is False)

    Returns
    -------
    y : numpy array
        array with shape specified by :attr:`same`.
    """

    nfft, Nx, Nh = np.int32([nfft, Nx, Nh])
    N = Nx + Nh - 1
    Nextra = nfft - N

    if nfft < N:
        raise ValueError("~~~To get right results, nfft must be larger than Nx+Nh-1!")

    if ftshift:
        if Nextra == 0:
            if np.mod(Nx, 2) == 0 and np.mod(Nh, 2) > 0:
                y = cut(y, ((0, N), ), axis)
            else:
                y = cut(y, ((1, nfft), (0, 1)), axis)
        else:
            if np.mod(Nx, 2) == 0 and np.mod(Nextra, 2) == 0:
                Nhead = np.int32(np.fix(Nextra / 2.))
            else:
                Nhead = np.int32(np.fix(Nextra / 2.) + 1)
            Nstart2 = Nhead
            Nend2 = np.int32(Nstart2 + N)
            y = cut(y, ((Nstart2, Nend2), ), axis)
    else:
        Nstart2 = 0
        Nend2 = Nx
        Nend1 = nfft
        Nstart1 = int(np.uint(Nend1 - (Nh - 1)))
        y = cut(y, ((Nstart1, Nend1), (Nstart2, Nend2)), axis)

    if shape in ['same', 'SAME', 'Same']:
        Nstart = np.uint(np.fix(Nh / 2.))
        Nend = np.uint(Nstart + Nx)
    elif shape in ['valid', 'VALID', 'Valid']:
        Nstart = np.uint(Nh - 1)
        Nend = np.uint(N - (Nh - 1))
    elif shape in ['full', 'FULL', 'Full']:
        Nstart, Nend = (0, N)
    y = cut(y, ((Nstart, Nend),), axis=axis)
    return y


def fftcorr1(x, h, shape='same', axis=0, nfft=None, ftshift=False, eps=None):
    """Correlation using Fast Fourier Transformation

    Correlation using Fast Fourier Transformation.

    Parameters
    ----------
    x : numpy array
        data to be convolved.
    h : numpy array
        filter array
    shape : str, optional
        output shape:
        1. ``'same'`` --> same size as input x, :math:`N_x`
        2. ``'valid'`` --> valid correlation output
        3. ``'full'`` --> full correlation output, :math:`N_x+N_h-1`
        (the default is 'same')
    axis : int, optional
        correlation axis (the default is 0)
    nfft : int, optional
        number of fft points (the default is None, :math:`2^{nextpow2(N_x+N_h-1)}`),
        note that :attr:`nfft` can not be smaller than :math:`N_x+N_h-1`.
    ftshift : bool, optional
        whether shift frequencies (the default is False)
    eps : None or float, optional
        x[abs(x)<eps] = 0 (the default is None, does nothing)

    Returns
    -------
    y : numpy array
        Correlation result array.

    """

    if np.ndim(h) == 1 and axis > 0:
        h = np.expand_dims(h, list(range(axis)))
    Nh = np.size(h, axis)
    Nx = np.size(x, axis)

    N = Nx + Nh - 1
    if nfft is None:
        nfft = 2**nextpow2(N)
    else:
        if nfft < N:
            raise ValueError("~~~To get right results, nfft must be larger than Nx+Nh-1!")

    x = padfft(x, nfft, axis, ftshift)
    h = padfft(h, nfft, axis, ftshift)
    X = fft(x, nfft, axis, norm=None, shift=ftshift)
    H = fft(h, nfft, axis, norm=None, shift=ftshift)
    Y = X * np.conj(H)
    y = ifft(Y, nfft, axis, norm=None, shift=ftshift)

    y = cutfftcorr1(y, nfft, Nx, Nh, shape, axis, ftshift)

    if eps is not None:
        y[abs(y) < eps] = 0.

    return y


def xcorr(A, B, shape='same', mod=None, axis=0):
    r"""Cross-correlation function estimates.


    Parameters
    ----------
    A : numpy array
        data1
    B : numpy array
        data2
    shape : str, optional
        output shape:
        1. ``'same'`` --> same size as input x, :math:`N_x`
        2. ``'valid'`` --> valid correlation output
        3. ``'full'`` --> full correlation output, :math:`N_x+N_h-1`
    mod : str, optional
        - ``'biased'``: scales the raw cross-correlation by 1/M.
        - ``'unbiased'``: scales the raw correlation by 1/(M-abs(lags)).
        - ``'coeff'``: normalizes the sequence so that the auto-correlations
                   at zero lag are identically 1.0.
        - :obj:`None`: no scaling (this is the default).
    """

    if np.ndim(A) == 1 and np.ndim(B) == 1:
        Ma, Mb = (1, 1)
        Na, Nb = (len(A), len(B))
    if np.ndim(A) == 2 and np.ndim(B) == 2:
        print(A.shape, B.shape)
        Ma, Na = A.shape
        Mb, Nb = B.shape
        if axis == 1 and Ma != Mb:
            raise ValueError("~~~Array A and B should have the same rows!")
        if axis == 0 and Na != Nb:
            raise ValueError("~~~Array A and B should have the same cols!")
    if shape in ['same', 'SAME']:
        Nc = max(Na, Nb)
    elif shape in ['full', 'FULL']:
        Nc = Na + Nb - 1
    elif shape in ['valid', 'VALID']:
        Nc = max(Na, Nb) - max(Na, Nb) + 1
    else:
        raise ValueError("~~~Not supported shape:" + shape + "!")

    CPLXDTYPESTR = ['complex128', 'complex64', 'complex']

    if A.dtype in CPLXDTYPESTR or B.dtype in CPLXDTYPESTR:
        dtype = 'complex'
    else:
        dtype = 'float'

    if np.ndim(A) == 1 and np.ndim(B) == 1:
        C = np.correlate(A, B, mode=shape)
    if np.ndim(A) == 2 and np.ndim(B) == 2:
        C = np.zeros((Ma, Nc), dtype=dtype)
        if axis == 0:
            for n in range(Na):
                C[:, n] = np.correlate(A[:, n], B[:, n], mode=shape)
        if axis == 1:
            for m in range(Ma):
                C[m, :] = np.correlate(A[m, :], B[m, :], mode=shape)
    return C


def accc(Sr, isplot=False):
    r"""Average cross correlation coefficient

    Average cross correlation coefficient (ACCC)

    .. math::
       \overline{C(\eta)}=\sum_{\eta} s^{*}(\eta) s(\eta+\Delta \eta)

    where, :math:`\eta, \Delta \eta` are azimuth time and it's increment.


    Parameters
    ----------
    Sr : numpy array
        SAR raw signal data :math:`N_a\times N_r` or range compressed data.

    Returns
    -------
    1d array
        ACCC in each range cell.
    """

    Na, Nr = Sr.shape

    acccv = np.sum(Sr[1:, :] * np.conj(Sr[0:-1, :]), 0)

    if isplot:
        plt.figure()
        plt.subplot(121)
        pyailib.cplot(acccv, '-b')
        plt.title('ACCC (all range cell)')
        plt.subplot(122)
        ccv = Sr[1:, 0] * np.conj(Sr[0:-1, 0])
        pyailib.cplot(ccv, '-b')
        pyailib.cplot([np.mean(ccv)], '-r')
        plt.title('CCC (0-th range cell)')
        plt.show()

    return acccv


if __name__ == '__main__':

    ftshift = False
    x = np.array([1, 2, 3, 4, 5])
    h = np.array([1 + 2j, 2, 3, 4, 5, 6, 7])

    y1 = corr1(x, h, shape='same')
    y2 = fftcorr1(x, h, axis=0, nfft=None, shape='same', ftshift=ftshift)
    # print(y1)
    # print(y2)
    print(np.sum(np.abs(y1 - y2)), np.sum(np.angle(y1) - np.angle(y2)))

    y1 = corr1(x, h, shape='valid')
    y2 = fftcorr1(x, h, axis=0, nfft=None, shape='valid', ftshift=ftshift)
    # print(y1)
    # print(y2)
    print(np.sum(np.abs(y1 - y2)), np.sum(np.angle(y1) - np.angle(y2)))

    y1 = corr1(x, h, shape='full')
    y2 = fftcorr1(x, h, axis=0, nfft=None, shape='full', ftshift=ftshift)
    # print(y1)
    # print(y2)
    print(np.sum(np.abs(y1 - y2)), np.sum(np.angle(y1) - np.angle(y2)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-07-06 09:32:16
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

import numpy as np
import pyailib as pl
import matplotlib.pyplot as plt


def plot_bbox(bboxes, labels=None, scores=None, edgecolors=None, linewidths=1, fontdict=None, textpos='TopCenter', offset=None, ax=None):
    r"""Plots bounding boxes with scores and labels


    Parameters
    ----------
    bboxes : list or numpy array
        The bounding boxes, in ``LeftTopRightBottom`` mode, which means (xmin, ymin, xmax, ymax)
    labels : list or None, optional
        The labels, can be a list of class id or class name. If None, won't show labels.
    scores : list or None, optional
        The scores, can be a list of float numbers. If None, won't show labels.
    edgecolors : None, optional
        The edgecolors for bounding boxes.
    linewidths : int, optional
        The linewidths for bounding boxes.
    fontdict : None, optional
        The fontdict for labels and scores.
    textpos : str, optional
        The position for text (labels and scores).
    offset : None, optional
        The offset for text (labels and scores).
    ax : None, optional
        The ``ax`` handle, If None, auto generated.

    Returns
    -------
    ax
        The ``ax`` handle

    see :func:`fmt_bbox`

    Example
    -------

    Plot bounding boxes with scores and labels on an image.

    .. image:: ./_static/demo_plot_bboxes.png
       :scale: 100 %
       :align: center

    The results shown in the above figure can be obtained by the following codes.

    ::

        import pyailib as pl
        import matplotlib.pyplot as plt

        bboxes = [[100, 100, 200, 200], [300, 300, 400, 500]]
        labels = ['dog', 'cat']
        scores = [0.987, None]
        edgecolors = [list(pl.DISTINCT_COLORS_RGB_NORM.values())[0], None]
        edgecolors = list(pl.DISTINCT_COLORS_RGB_NORM.values())[0:2]
        linewidths = [2, 4]

        fontdict = {'family': 'Times New Roman',
                    'style': 'italic',
                    'weight': 'normal',
                    'color': 'darkred',
                    'size': 12,
                    }

        x = pl.imread('../../data/images/LenaRGB512.tif')
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.imshow(x)

        pl.plot_bbox(bboxes, labels=labels, scores=scores, edgecolors=edgecolors, linewidths=linewidths, fontdict=fontdict, textpos='TopLeft', ax=ax)
        plt.axis('off')
        plt.savefig('./bbbox.png', bbox_inches='tight', pad_inches=0)
        plt.show()

    """
    nbox = len(bboxes)
    labels = [labels] * nbox if labels is None or type(labels) is str else labels
    labels = labels * nbox if len(labels) == 1 else labels
    scores = [scores] * nbox if scores is None or type(scores) is float else scores
    scores = scores * nbox if len(scores) == 1 else scores
    # edgecolors = pl.DISTINCT_COLORS['DistinctNormRGB20'][slice(0, nbox)] if edgecolors is None else edgecolors
    edgecolors = [edgecolors] * nbox if edgecolors is None else edgecolors
    edgecolors = edgecolors * nbox if len(edgecolors) == 1 else edgecolors
    edgecolors = [edgecolors] * nbox if type(edgecolors) is str or len(edgecolors) != nbox else edgecolors

    linewidths = [linewidths] * nbox if type(linewidths) is float or type(linewidths) is int or linewidths is None else linewidths
    linewidths = linewidths * nbox if len(linewidths) == 1 else linewidths
    offset = [6, 0] if offset is None else offset
    offset = [offset] * 2 if type(offset) is int or type(offset) is float else offset

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

    for bbox, label, score, edgecolor, linewidth in zip(bboxes, labels, scores, edgecolors, linewidths):
        xy = (bbox[1], bbox[0])
        height = bbox[2] - bbox[0]
        width = bbox[3] - bbox[1]
        ax.add_patch(plt.Rectangle(xy, width, height, fill=False, edgecolor=edgecolor, linewidth=linewidth))
        caption = ''
        if label is not None:
            caption += str(label)
        if score is not None:
            caption += ': {:.2f}'.format(score)

        if textpos in ['TopCenter', 'topcenter']:
            ax.text(bbox[1] + width / 2 - offset[1], bbox[0] - offset[0], caption, fontdict=fontdict)
        if textpos in ['BottomCenter', 'bottomcenter']:
            ax.text(bbox[1] + width / 2 - offset[1], bbox[2] + offset[0], caption, fontdict=fontdict)
        if textpos in ['TopLeft', 'topleft']:
            ax.text(bbox[1] - offset[1], bbox[0] - offset[0], caption, fontdict=fontdict)
        if textpos in ['BottomLeft', 'bottomleft']:
            ax.text(bbox[1] - offset[1], bbox[2] + offset[0], caption, fontdict=fontdict)
    return ax


# def fmt_bbox(bboxes, fmtstr='LTRB2CHW'):
#     r"""Formats bounding boxes

#     .. warning:: The height and width are computed by :math:`y_{\rm max} - y_{\rm min}` and :math:`x_{\rm max} - x_{\rm min}`.


#     Parameters
#     ----------
#     bboxes : list or numpy array
#         The bounding boxes to be converted, all bboxes have the same mode.
#     fmtstr : str, optional
#         - ``'LTRB2TLBR'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
#         - ``'TLBR2LTRB'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
#         - ``'CWH2CHW'``: CenterWidthHeight (x, y, w, h) --> CenterHeightWidth (y, x, h, w)
#         - ``'CHW2CWH'``: CenterHeightWidth (y, x, h, w) --> CenterWidthHeight (x, y, w, h)
#         - ``'LTRB2CWH'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> CenterWidthHeight (x, y, w, h)
#         - ``'LTRB2CHW'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> CenterHeightWidth (y, x, h, w)
#         - ``'TLBR2CWH'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> CenterWidthHeight (x, y, w, h)
#         - ``'TLBR2CHW'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> CenterHeightWidth (y, x, h, w)
#         - ``'CWH2LTRB'``: CenterWidthHeight (x, y, w, h) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
#         - ``'CWH2TLBR'``: CenterWidthHeight (x, y, w, h) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
#         - ``'CHW2LTRB'``: CenterHeightWidth (y, x, h, w) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
#         - ``'CHW2TLBR'``: CenterHeightWidth (y, x, h, w) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
#         - ``'LRTB2LTRB'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
#         - ``'LRTB2TLBR'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
#         - ``'LTRB2LRTB'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)
#         - ``'LRTB2CWH'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> CenterWidthHeight (x, y, w, h)
#         - ``'LRTB2CHW'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> CenterHeightWidth (y, x, h, w)
#         - ``'CWH2LRTB'``: CenterWidthHeight (x, y, w, h) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)
#         - ``'CHW2LRTB'``: CenterHeightWidth (y, x, h, w) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)

#     Returns
#     -------
#     list or numpy array
#         The formated bounding boxes.

#     see :func:`plot_bbox`

#     """
#     out = []
#     if fmtstr in ['LTRB2TLBR', 'LeftTopRightBottom2TopLeftBottomRight']:
#         for bbox in bboxes:
#             xmin, ymin, xmax, ymax = bbox
#             out.append([ymin, xmin, ymax, xmax])
#     if fmtstr in ['TLBR2LTRB', 'TopLeftBottomRight2LeftTopRightBottom']:
#         for bbox in bboxes:
#             ymin, xmin, ymax, xmax = bbox
#             out.append([xmin, ymin, xmax, ymax])
#     if fmtstr in ['CWH2CHW', 'CenterWidthHeight2CenterHeightWidth']:
#         for bbox in bboxes:
#             x, y, w, h = bbox
#             out.append([y, x, h, w])
#     if fmtstr in ['CHW2CWH', 'CenterHeightWidth2CenterWidthHeight']:
#         for bbox in bboxes:
#             y, x, h, w = bbox
#             out.append([x, y, w, h])
#     if fmtstr in ['LTRB2CWH', 'LeftTopRightBottom2CenterWidthHeight']:
#         for bbox in bboxes:
#             xmin, ymin, xmax, ymax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([x, y, w, h])
#     if fmtstr in ['LTRB2CHW', 'LeftTopRightBottom2CenterHeightWidth']:
#         for bbox in bboxes:
#             xmin, ymin, xmax, ymax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([y, x, h, w])
#     if fmtstr in ['TLBR2CWH', 'LeftTopRightBottom2CenterWidthHeight']:
#         for bbox in bboxes:
#             ymin, xmin, ymax, xmax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([x, y, w, h])
#     if fmtstr in ['TLBR2CHW', 'LeftTopRightBottom2CenterHeightWidth']:
#         for bbox in bboxes:
#             ymin, xmin, ymax, xmax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([y, x, h, w])
#     if fmtstr in ['CWH2LTRB', 'CenterWidthHeight2LeftTopRightBottom']:
#         for bbox in bboxes:
#             x, y, w, h = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([xmin, ymin, xmax, ymax])
#     if fmtstr in ['CWH2TLBR', 'CenterWidthHeight2TopLeftBottomRight']:
#         for bbox in bboxes:
#             x, y, w, h = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([ymin, xmin, ymax, xmax])
#     if fmtstr in ['CHW2LTRB', 'CenterHeightWidth2LeftTopRightBottom']:
#         for bbox in bboxes:
#             y, x, h, w = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([xmin, ymin, xmax, ymax])
#     if fmtstr in ['CHW2TLBR', 'CenterHeightWidth2TopLeftBottomRight']:
#         for bbox in bboxes:
#             y, x, h, w = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([ymin, xmin, ymax, xmax])
#     if fmtstr in ['LRTB2LTRB', 'LeftRightTopBottom2LeftTopRightBottom']:
#         for bbox in bboxes:
#             xmin, xmax, ymin, ymax = bbox
#             out.append([xmin, ymin, xmax, ymax])
#     if fmtstr in ['LRTB2TLBR', 'LeftRightTopBottom2TopLeftBottomRight']:
#         for bbox in bboxes:
#             xmin, xmax, ymin, ymax = bbox
#             out.append([ymin, xmin, ymax, xmax])
#     if fmtstr in ['LTRB2LRTB', 'LeftTopRightBottom2LeftRightTopBottom']:
#         for bbox in bboxes:
#             xmin, ymin, xmax, ymax = bbox
#             out.append([xmin, xmax, ymin, ymax])
#     if fmtstr in ['TLBR2LRTB', 'TopLeftBottomRight2LeftRightTopBottom']:
#         for bbox in bboxes:
#             ymin, xmin, ymax, xmax = bbox
#             out.append([xmin, xmax, ymin, ymax])
#     if fmtstr in ['LRTB2CWH', 'LeftRightTopBottom2CenterWidthHeight']:
#         for bbox in bboxes:
#             xmin, xmax, ymin, ymax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([x, y, w, h])
#     if fmtstr in ['LRTB2CHW', 'LeftRightTopBottom2CenterHeightWidth']:
#         for bbox in bboxes:
#             xmin, xmax, ymin, ymax = bbox
#             x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
#             w, h = xmax - xmin, ymax - ymin
#             out.append([y, x, h, w])
#     if fmtstr in ['CWH2LRTB', 'CenterWidthHeight2LeftRightTopBottom']:
#         for bbox in bboxes:
#             x, y, w, h = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([xmin, xmax, ymin, ymax])
#     if fmtstr in ['CHW2LRTB', 'CenterHeightWidth2LeftRightTopBottom']:
#         for bbox in bboxes:
#             y, x, h, w = bbox
#             xmin, xmax = x - w / 2., x + w / 2.
#             ymin, ymax = y - h / 2., y + h / 2.
#             out.append([xmin, xmax, ymin, ymax])
#     if type(bboxes) is np.ndarray:
#         return np.array(out)
#     else:
#         return out


def fmt_bbox(bboxes, fmtstr='LTRB2CHW'):
    r"""Formats bounding boxes

    .. warning:: The height and width are computed by :math:`y_{\rm max} - y_{\rm min}` and :math:`x_{\rm max} - x_{\rm min}`.


    Parameters
    ----------
    bboxes : list or numpy array
        The bounding boxes to be converted, all bboxes have the same mode.
    fmtstr : str, optional
        - ``'LTRB2TLBR'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
        - ``'TLBR2LTRB'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
        - ``'CWH2CHW'``: CenterWidthHeight (x, y, w, h) --> CenterHeightWidth (y, x, h, w)
        - ``'CHW2CWH'``: CenterHeightWidth (y, x, h, w) --> CenterWidthHeight (x, y, w, h)
        - ``'LTRB2CWH'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> CenterWidthHeight (x, y, w, h)
        - ``'LTRB2CHW'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> CenterHeightWidth (y, x, h, w)
        - ``'TLBR2CWH'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> CenterWidthHeight (x, y, w, h)
        - ``'TLBR2CHW'``: TopLeftBottomRight (ymin, xmin, ymax, xmax) --> CenterHeightWidth (y, x, h, w)
        - ``'CWH2LTRB'``: CenterWidthHeight (x, y, w, h) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
        - ``'CWH2TLBR'``: CenterWidthHeight (x, y, w, h) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
        - ``'CHW2LTRB'``: CenterHeightWidth (y, x, h, w) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
        - ``'CHW2TLBR'``: CenterHeightWidth (y, x, h, w) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
        - ``'LRTB2LTRB'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> LeftTopRightBottom (xmin, ymin, xmax, ymax)
        - ``'LRTB2TLBR'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> TopLeftBottomRight (ymin, xmin, ymax, xmax)
        - ``'LTRB2LRTB'``: LeftTopRightBottom (xmin, ymin, xmax, ymax) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)
        - ``'LRTB2CWH'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> CenterWidthHeight (x, y, w, h)
        - ``'LRTB2CHW'``: LeftRightTopBottom (xmin, xmax, ymin, ymax) --> CenterHeightWidth (y, x, h, w)
        - ``'CWH2LRTB'``: CenterWidthHeight (x, y, w, h) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)
        - ``'CHW2LRTB'``: CenterHeightWidth (y, x, h, w) --> LeftRightTopBottom (xmin, xmax, ymin, ymax)

    Returns
    -------
    list or numpy array
        The formated bounding boxes.

    see :func:`plot_bbox`

    """

    dtype = type(bboxes)
    bboxes = np.array(bboxes, dtype=np.float32)
    out = np.copy(bboxes)
    if fmtstr in ['LTRB2TLBR', 'LeftTopRightBottom2TopLeftBottomRight']:
        # xmin, ymin, xmax, ymax = bbox
        # out.append([ymin, xmin, ymax, xmax])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 1], bboxes[:, 0], bboxes[:, 3], bboxes[:, 2]
    if fmtstr in ['TLBR2LTRB', 'TopLeftBottomRight2LeftTopRightBottom']:
        # ymin, xmin, ymax, xmax = bbox
        # out.append([xmin, ymin, xmax, ymax])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 1], bboxes[:, 0], bboxes[:, 3], bboxes[:, 2]
    if fmtstr in ['CWH2CHW', 'CenterWidthHeight2CenterHeightWidth']:
        # x, y, w, h = bbox
        # out.append([y, x, h, w])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 1], bboxes[:, 0], bboxes[:, 3], bboxes[:, 2]
    if fmtstr in ['CHW2CWH', 'CenterHeightWidth2CenterWidthHeight']:
        # y, x, h, w = bbox
        # out.append([x, y, w, h])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 1], bboxes[:, 0], bboxes[:, 3], bboxes[:, 2]
    if fmtstr in ['LTRB2CWH', 'LeftTopRightBottom2CenterWidthHeight']:
        # xmin, ymin, xmax, ymax = bbox
        # x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
        # w, h = xmax - xmin, ymax - ymin
        # out.append([x, y, w, h])
        out[:, 0], out[:, 1] = (bboxes[:, 0] + bboxes[:, 2]) / 2., (bboxes[:, 1] + bboxes[:, 3]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 2] - bboxes[:, 0], bboxes[:, 3] - bboxes[:, 1]
    if fmtstr in ['LTRB2CHW', 'LeftTopRightBottom2CenterHeightWidth']:
        # xmin, ymin, xmax, ymax = bbox
        # y, x = (ymin + ymax) / 2., (xmin + xmax) / 2.
        # h, w = ymax - ymin, xmax - xmin
        # out.append([y, x, h, w])
        out[:, 0], out[:, 1] = (bboxes[:, 1] + bboxes[:, 3]) / 2., (bboxes[:, 0] + bboxes[:, 2]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 3] - bboxes[:, 1], bboxes[:, 2] - bboxes[:, 0]
    if fmtstr in ['TLBR2CWH', 'LeftTopRightBottom2CenterWidthHeight']:
        # ymin, xmin, ymax, xmax = bbox
        # x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
        # w, h = xmax - xmin, ymax - ymin
        # out.append([x, y, w, h])
        out[:, 0], out[:, 1] = (bboxes[:, 1] + bboxes[:, 3]) / 2., (bboxes[:, 0] + bboxes[:, 2]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 3] - bboxes[:, 1], bboxes[:, 2] - bboxes[:, 0]
    if fmtstr in ['TLBR2CHW', 'LeftTopRightBottom2CenterHeightWidth']:
        # ymin, xmin, ymax, xmax = bbox
        # y, x = (ymin + ymax) / 2., (xmin + xmax) / 2.
        # h, w = ymax - ymin, xmax - xmin
        # out.append([y, x, h, w])
        out[:, 0], out[:, 1] = (bboxes[:, 0] + bboxes[:, 2]) / 2., (bboxes[:, 1] + bboxes[:, 3]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 2] - bboxes[:, 0], bboxes[:, 3] - bboxes[:, 1]
    if fmtstr in ['CWH2LTRB', 'CenterWidthHeight2LeftTopRightBottom']:
        # x, y, w, h = bbox
        # xmin, ymin = x - w / 2., y - h / 2.
        # xmax, ymax = x + w / 2., y + h / 2.
        # out.append([xmin, ymin, xmax, ymax])
        out[:, 0], out[:, 1] = bboxes[:, 0] - bboxes[:, 2] / 2., bboxes[:, 1] - bboxes[:, 3] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 0] + bboxes[:, 2] / 2., bboxes[:, 1] + bboxes[:, 3] / 2.
    if fmtstr in ['CWH2TLBR', 'CenterWidthHeight2TopLeftBottomRight']:
        # x, y, w, h = bbox
        # ymin, xmin = y - h / 2., x - w / 2.
        # ymax, xmax = y + h / 2., x + w / 2.
        # out.append([ymin, xmin, ymax, xmax])
        out[:, 0], out[:, 1] = bboxes[:, 1] - bboxes[:, 3] / 2., bboxes[:, 0] - bboxes[:, 2] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 1] + bboxes[:, 3] / 2., bboxes[:, 0] + bboxes[:, 2] / 2.
    if fmtstr in ['CHW2LTRB', 'CenterHeightWidth2LeftTopRightBottom']:
        # y, x, h, w = bbox
        # xmin, ymin = x - w / 2., y - h / 2.
        # xmax, ymax = x + w / 2., y + h / 2.
        # out.append([xmin, ymin, xmax, ymax])
        out[:, 0], out[:, 1] = bboxes[:, 1] - bboxes[:, 3] / 2., bboxes[:, 0] - bboxes[:, 2] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 1] + bboxes[:, 3] / 2., bboxes[:, 0] + bboxes[:, 2] / 2.
    if fmtstr in ['CHW2TLBR', 'CenterHeightWidth2TopLeftBottomRight']:
        # y, x, h, w = bbox
        # ymin, xmin = y - h / 2., x - w / 2.
        # ymax, xmax = y + h / 2., x + w / 2.
        # out.append([ymin, xmin, ymax, xmax])
        out[:, 0], out[:, 1] = bboxes[:, 0] - bboxes[:, 2] / 2., bboxes[:, 1] - bboxes[:, 3] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 0] + bboxes[:, 2] / 2., bboxes[:, 1] + bboxes[:, 3] / 2.
    if fmtstr in ['LRTB2LTRB', 'LeftRightTopBottom2LeftTopRightBottom']:
        # xmin, xmax, ymin, ymax = bbox
        # out.append([xmin, ymin, xmax, ymax])
        out[:, 1], out[:, 2] = bboxes[:, 2], bboxes[:, 1]
    if fmtstr in ['LRTB2TLBR', 'LeftRightTopBottom2TopLeftBottomRight']:
        # xmin, xmax, ymin, ymax = bbox
        # out.append([ymin, xmin, ymax, xmax])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 2], bboxes[:, 0], bboxes[:, 3], bboxes[:, 1]
    if fmtstr in ['LTRB2LRTB', 'LeftTopRightBottom2LeftRightTopBottom']:
        # xmin, ymin, xmax, ymax = bbox
        # out.append([xmin, xmax, ymin, ymax])
        out[:, 1], out[:, 2] = bboxes[:, 2], bboxes[:, 1]
    if fmtstr in ['TLBR2LRTB', 'TopLeftBottomRight2LeftRightTopBottom']:
        # ymin, xmin, ymax, xmax = bbox
        # out.append([xmin, xmax, ymin, ymax])
        out[:, 0], out[:, 1], out[:, 2], out[:, 3] = bboxes[:, 1], bboxes[:, 3], bboxes[:, 0], bboxes[:, 2]
    if fmtstr in ['LRTB2CWH', 'LeftRightTopBottom2CenterWidthHeight']:
        # xmin, xmax, ymin, ymax = bbox
        # x, y = (xmin + xmax) / 2., (ymin + ymax) / 2.
        # w, h = xmax - xmin, ymax - ymin
        # out.append([x, y, w, h])
        out[:, 0], out[:, 1] = (bboxes[:, 0] + bboxes[:, 1]) / 2., (bboxes[:, 2] + bboxes[:, 3]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 1] - bboxes[:, 0], bboxes[:, 3] - bboxes[:, 2]
    if fmtstr in ['LRTB2CHW', 'LeftRightTopBottom2CenterHeightWidth']:
        # xmin, xmax, ymin, ymax = bbox
        # y, x = (ymin + ymax) / 2., (xmin + xmax) / 2.
        # h, w = ymax - ymin, xmax - xmin
        # out.append([y, x, h, w])
        out[:, 0], out[:, 1] = (bboxes[:, 2] + bboxes[:, 3]) / 2., (bboxes[:, 0] + bboxes[:, 1]) / 2.
        out[:, 2], out[:, 3] = bboxes[:, 3] - bboxes[:, 2], bboxes[:, 1] - bboxes[:, 0]
    if fmtstr in ['CWH2LRTB', 'CenterWidthHeight2LeftRightTopBottom']:
        # x, y, w, h = bbox
        # xmin, xmax = x - w / 2., x + w / 2.
        # ymin, ymax = y - h / 2., y + h / 2.
        # out.append([xmin, xmax, ymin, ymax])
        out[:, 0], out[:, 1] = bboxes[:, 0] - bboxes[:, 2] / 2., bboxes[:, 0] + bboxes[:, 2] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 1] - bboxes[:, 3] / 2., bboxes[:, 1] + bboxes[:, 3] / 2.
    if fmtstr in ['CHW2LRTB', 'CenterHeightWidth2LeftRightTopBottom']:
        # y, x, h, w = bbox
        # xmin, xmax = x - w / 2., x + w / 2.
        # ymin, ymax = y - h / 2., y + h / 2.
        # out.append([xmin, xmax, ymin, ymax])
        out[:, 0], out[:, 1] = bboxes[:, 1] - bboxes[:, 3] / 2., bboxes[:, 1] + bboxes[:, 3] / 2.
        out[:, 2], out[:, 3] = bboxes[:, 0] - bboxes[:, 2] / 2., bboxes[:, 0] + bboxes[:, 2] / 2.

    if dtype is list:
        return out.tolist()
    else:
        return out


if __name__ == '__main__':

    bboxes = [[100, 100, 200, 200], [300, 300, 400, 500]]
    labels = ['dog', 'cat']
    scores = [0.987, None]
    edgecolors = [list(pl.DISTINCT_COLORS_RGB_NORM.values())[0], None]
    edgecolors = list(pl.DISTINCT_COLORS_RGB_NORM.values())[0:2]
    linewidths = 1

    fontdict = {'family': 'Times New Roman',
                'style': 'italic',
                'weight': 'normal',
                'color': 'darkred',
                'size': 16,
                }

    x = pl.imread('../../data/images/LenaRGB512.tif')
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(x)

    plot_bbox(bboxes, labels=labels, scores=scores, edgecolors=edgecolors, linewidths=linewidths, textpos='TopLeft', fontdict=fontdict, ax=ax)

    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-03-13 10:34:43
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

# DEFINE GLOBAL CONSTS
C = 299792458.0  # 2.99792458e8 in m/s
# PI = 3.14159265358979323846
PI = 3.141592653589793238462643383279502884197169399375105820974944592307
EPS = 1.0e-32


Ge = 398603e9  # the constant of earth gravitation
Rea = 6367.856e3  # the avarage radius of earth
Ree = 6378.137e3  # the equatorial radius of earth
Rep = 6356.752e3  # the polar radius of earth



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-06 21:14:04
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import numpy as np
import pyailib as pl


def ampphaerror(orig, reco):
    r"""compute amplitude phase error

    compute amplitude phase error of two complex valued matrix

    Parameters
    ----------
    orig : numpy array
        orignal
    reco : numpy array
        reconstructed

    Returns
    -------
    amperror : float
        error of amplitude
    phaerror : float
        error of phase
    """

    amp_orig = np.abs(orig)
    amp_reco = np.abs(reco)
    pha_orig = np.angle(orig)
    pha_reco = np.angle(reco)
    amperror = np.mean(np.abs(amp_orig - amp_reco))
    phaerror = np.mean(np.abs(pha_orig - pha_reco))

    return amperror, phaerror


def mse(X, Y, caxis=None, axis=None, norm=False, reduction='mean'):
    r"""computes the mean square error

    Both complex and real representation are supported.

    .. math::
       {\rm MSE}({\bf X, Y}) = \frac{1}{N}\|{\bf X} - {\bf Y}\|_2^2 = \frac{1}{N}\sum_{i=1}^N(|x_i - y_i|)^2

    Parameters
    ----------
    X : array
        reconstructed
    Y : array
        target
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    norm : bool
        If :obj:`True`, normalize with the f-norm of :attr:`X` and :attr:`Y`. (default is :obj:`False`)
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    scalar or array
         mean square error

    Examples
    ---------

    ::

        norm = False
        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)
        Y = np.random.randn(5, 2, 3, 4)

        # real
        C1 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
        C1 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # ---output
        [[1.57602573 2.32844311]
        [1.07232374 2.36118382]
        [2.1841515  0.79002805]
        [2.43036295 3.18413899]
        [2.31107373 2.73990485]] 20.977636476183186 2.0977636476183186
        [3.90446884 3.43350757 2.97417955 5.61450194 5.05097858] 20.977636476183186 4.195527295236637
        [3.90446884 3.43350757 2.97417955 5.61450194 5.05097858] 20.977636476183186 4.195527295236637

    """

    X = X - Y
    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
            E = np.mean((X.conj() * X).real)
        else:
            E = np.mean((X.conj() * X).real, axis=axis)
    else:
        if caxis is None:  # real
            if axis is None:
               E = np.mean(X**2)
            else:
               E = np.mean(X**2, axis=axis)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
               E = np.mean(X[idxreal]**2 + X[idximag]**2)
            else:
               E = np.mean(X[idxreal]**2 + X[idximag]**2, axis=axis)

    if norm is True:
        xnorm = pl.fnorm(X, caxis=caxis, axis=axis, reduction=None)
        ynorm = pl.fnorm(Y, caxis=caxis, axis=axis, reduction=None)
        E /= (xnorm * ynorm + pl.EPS)
        
    if reduction in ['mean', 'MEAN']:
       E = np.mean(E)
    if reduction in ['sum', 'SUM']:
       E = np.sum(E)

    return E


def sse(X, Y, caxis=None, axis=None, norm=False, reduction='mean'):
    r"""computes the sum square error

    Both complex and real representation are supported.

    .. math::
       {\rm SSE}({\bf X, Y}) = \|{\bf X} - {\bf Y}\|_2^2 = \sum_{i=1}^N(|x_i - y_i|)^2

    Parameters
    ----------
    X : array
        reconstructed
    Y : array
        target
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    norm : bool
        If :obj:`True`, normalize with the f-norm of :attr:`X` and :attr:`Y`. (default is :obj:`False`)
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    scalar or array
         sum square error

    Examples
    ---------

    ::

        norm = False
        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)
        Y = np.random.randn(5, 2, 3, 4)

        # real
        C1 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
        C1 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # ---output
        [[18.91230872 27.94131733]
        [12.86788492 28.33420589]
        [26.209818    9.48033663]
        [29.16435541 38.20966786]
        [27.73288477 32.87885818]] 251.73163771419823 25.173163771419823
        [46.85362605 41.20209081 35.69015462 67.37402327 60.61174295] 251.73163771419823 50.346327542839646
        [46.85362605 41.20209081 35.69015462 67.37402327 60.61174295] 251.73163771419823 50.346327542839646

    """

    X = X - Y
    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
            E = np.sum((X.conj() * X).real)
        else:
            E = np.sum((X.conj() * X).real, axis=axis)
    else:
        if caxis is None:  # real
            if axis is None:
               E = np.sum(X**2)
            else:
               E = np.sum(X**2, axis=axis)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
               E = np.sum(X[idxreal]**2 + X[idximag]**2)
            else:
               E = np.sum(X[idxreal]**2 + X[idximag]**2, axis=axis)

    if norm is True:
        xnorm = pl.fnorm(X, caxis=caxis, axis=axis, reduction=None)
        ynorm = pl.fnorm(Y, caxis=caxis, axis=axis, reduction=None)
        E /= (xnorm * ynorm + pl.EPS)

    if reduction in ['mean', 'MEAN']:
       E = np.mean(E)
    if reduction in ['sum', 'SUM']:
       E = np.sum(E)

    return E


def mae(X, Y, caxis=None, axis=None, norm=False, reduction='mean'):
    r"""computes the mean absoluted error

    Both complex and real representation are supported.

    .. math::
       {\rm MAE}({\bf X, Y}) = \frac{1}{N}\||{\bf X} - {\bf Y}|\| = \frac{1}{N}\sum_{i=1}^N |x_i - y_i|

    Parameters
    ----------
    X : array
        original
    X : array
        reconstructed
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    norm : bool
        If :obj:`True`, normalize with the f-norm of :attr:`X` and :attr:`Y`. (default is :obj:`False`)
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    scalar or array
         mean absoluted error

    Examples
    ---------

    ::

        norm = False
        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)
        Y = np.random.randn(5, 2, 3, 4)

        # real
        C1 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
        C1 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # ---output
        [[1.06029116 1.19884877]
        [0.90117091 1.13552361]
        [1.23422083 0.75743914]
        [1.16127965 1.42169262]
        [1.25090731 1.29134222]] 11.41271620974502 1.141271620974502
        [1.71298566 1.50327364 1.53328572 2.11430946 2.01435599] 8.878210471231741 1.7756420942463482
        [1.71298566 1.50327364 1.53328572 2.11430946 2.01435599] 8.878210471231741 1.7756420942463482

    """

    X = X - Y
    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
           E = np.mean(np.abs(X))
        else:
           E = np.mean(np.abs(X), axis=axis)
    else:
        if caxis is None:  # real
            if axis is None:
               E = np.mean(np.abs(X))
            else:
               E = np.mean(np.abs(X), axis=axis)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
               E = np.mean(np.sqrt(X[idxreal]**2 + X[idximag]**2))
            else:
               E = np.mean(np.sqrt(X[idxreal]**2 + X[idximag]**2), axis=axis)
    
    if norm is True:
        xnorm = pl.fnorm(X, caxis=caxis, axis=axis, reduction=None)
        ynorm = pl.fnorm(Y, caxis=caxis, axis=axis, reduction=None)
        E /= (xnorm * ynorm + pl.EPS)

    if reduction in ['mean', 'MEAN']:
       E = np.mean(E)
    if reduction in ['sum', 'SUM']:
       E = np.sum(E)

    return E


def sae(X, Y, caxis=None, axis=None, norm=False, reduction='mean'):
    r"""computes the sum absoluted error

    Both complex and real representation are supported.

    .. math::
       {\rm SAE}({\bf X, Y}) = \||{\bf X} - {\bf Y}|\| = \sum_{i=1}^N |x_i - y_i|

    Parameters
    ----------
    X : array
        original
    X : array
        reconstructed
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    norm : bool
        If :obj:`True`, normalize with the f-norm of :attr:`X` and :attr:`Y`. (default is :obj:`False`)
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    scalar or array
         sum absoluted error

    Examples
    ---------

    ::

        norm = False
        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)
        Y = np.random.randn(5, 2, 3, 4)

        # real
        C1 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
        C1 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
        C2 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
        C3 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
        print(C1, C2, C3)

        # ---output
        [[12.72349388 14.3861852 ]
        [10.81405096 13.62628335]
        [14.81065     9.08926963]
        [13.93535577 17.0603114 ]
        [15.0108877  15.49610662]] 136.95259451694022 13.695259451694023
        [20.55582795 18.03928365 18.39942858 25.37171356 24.17227192] 106.53852565478087 21.307705130956172
        [20.55582795 18.03928365 18.39942858 25.37171356 24.17227192] 106.5385256547809 21.30770513095618

    """

    X = X - Y
    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
           E = np.sum(np.abs(X))
        else:
           E = np.sum(np.abs(X), axis=axis)
    else:
        if caxis is None:  # real
            if axis is None:
               E = np.sum(np.abs(X))
            else:
               E = np.sum(np.abs(X), axis=axis)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
               E = np.sum(np.sqrt(X[idxreal]**2 + X[idximag]**2))
            else:
               E = np.sum(np.sqrt(X[idxreal]**2 + X[idximag]**2), axis=axis)

    if norm is True:
        xnorm = pl.fnorm(X, caxis=caxis, axis=axis, reduction=None)
        ynorm = pl.fnorm(Y, caxis=caxis, axis=axis, reduction=None)
        E /= (xnorm * ynorm + pl.EPS)

    if reduction in ['mean', 'MEAN']:
       E = np.mean(E)
    if reduction in ['sum', 'SUM']:
       E = np.sum(E)

    return E


if __name__ == '__main__':

    norm = False
    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)
    Y = np.random.randn(5, 2, 3, 4)

    # real
    C1 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
    C1 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)
    Y = np.random.randn(5, 2, 3, 4)

    # real
    C1 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sse(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
    C1 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sse(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)
    Y = np.random.randn(5, 2, 3, 4)

    # real
    C1 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
    C1 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = mae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)
    Y = np.random.randn(5, 2, 3, 4)

    # real
    C1 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sae(X, Y, caxis=1, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    Y = Y[:, 0, ...] + 1j * Y[:, 1, ...]
    C1 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction=None)
    C2 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='sum')
    C3 = sae(X, Y, caxis=None, axis=(-2, -1), norm=norm, reduction='mean')
    print(C1, C2, C3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-07-24 18:29:48
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import numpy as np
from pyailib.utils.const import EPS


def contrast(X, caxis=None, axis=None, mode='way1', reduction='mean'):
    r"""Compute contrast of an complex image

    ``'way1'`` is defined as follows, see [1]:

    .. math::
       C = \frac{\sqrt{{\rm E}\left(|I|^2 - {\rm E}(|I|^2)\right)^2}}{{\rm E}(|I|^2)}


    ``'way2'`` is defined as follows, see [2]:

    .. math::
        C = \frac{{\rm E}(|I|^2)}{\left({\rm E}(|I|)\right)^2}

    [1] Efficient Nonparametric ISAR Autofocus Algorithm Based on Contrast Maximization and Newton
    [2] section 13.4.1 in "Ian G. Cumming's SAR book"

    Parameters
    ----------
    X : numpy ndarray
        The image array.
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing contrast. The default is :obj:`None`, which means all. 
    mode : str, optional
        ``'way1'`` or ``'way2'``
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)

    Returns
    -------
    C : scalar or numpy array
        The contrast value of input.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)

        # real
        C1 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction=None)
        C2 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='sum')
        C3 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction=None)
        C2 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction='sum')
        C3 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        C1 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction=None)
        C2 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='sum')
        C3 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='mean')
        print(C1, C2, C3)

        # ---output
        [[1.07323512 1.39704055]
        [0.96033633 1.35878254]
        [1.57174342 1.42973702]
        [1.37236497 1.2351262 ]
        [1.06519696 1.4606771 ]] 12.924240207170865 1.2924240207170865
        [0.86507341 1.03834259 1.00448054 0.89381925 1.20616657] 5.007882367336851 1.0015764734673702
        [0.86507341 1.03834259 1.00448054 0.89381925 1.20616657] 5.007882367336851 1.0015764734673702
    """

    if np.iscomplex(X).any():  # complex in complex
        X = (X * X.conj()).real
    else:
        if caxis is None:  # real
            X = X**2
        else:  # complex in real
            X = np.sum(X**2, axis=caxis)

    if mode in ['way1', 'WAY1']:
        Xmean = X.mean(axis=axis, keepdims=True)
        C = np.sqrt(np.power(X - Xmean, 2).mean(axis=axis, keepdims=True)) / (Xmean + EPS)
        C = C.squeeze(axis)
    if mode in ['way2', 'WAY2']:
        C = X.mean(axis=axis) / (np.power((np.sqrt(X).mean(axis=axis, keepdims=True)), 2) + EPS)
        C = C.squeeze(axis)

    if reduction in ['mean', 'MEAN']:
        C = np.mean(C)
    if reduction in ['sum', 'SUM']:
        C = np.sum(C)
    return C


if __name__ == '__main__':

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)

    # real
    C1 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction=None)
    C2 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='sum')
    C3 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction=None)
    C2 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction='sum')
    C3 = contrast(X, caxis=1, axis=(-2, -1), mode='way1', reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    C1 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction=None)
    C2 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='sum')
    C3 = contrast(X, caxis=None, axis=(-2, -1), mode='way1', reduction='mean')
    print(C1, C2, C3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-23 07:01:55
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$


from __future__ import division, print_function, absolute_import
import os
import re
from pyailib.utils.convert import str2sec


def listxfile(listdir=None, exts=None, recursive=False, filelist=[]):
    r"""List the files in a directory.


    Parameters
    ----------
    listdir : None, optional
        The directory for listing. The default is None, which means return :attr:`filelist` directly.
    exts : str, list or None, optional
        File extension string or extension string list, such as ``'.png'`` or ``['.png', 'jpg']``.
        The default is None, which means any extension.
    recursive : bool, optional
        Recursive search? The default is False, which means only list the root directory (:attr:`listdir`).
    filelist : list, optional
        An initial list contains files's path. The default is ``[]``.

    Returns
    -------
    list
        The list of file path with extension :attr:`exts`. Sometimes you need to sort the list using ``sorted``. 

    """

    if listdir is None:
        return filelist

    exts = [exts] if type(exts) is str else exts

    for s in os.listdir(listdir):
        newDir = os.path.join(listdir, s)
        if os.path.isfile(newDir):
            if exts is not None:
                if newDir and(os.path.splitext(newDir)[1] in exts):
                    filelist.append(newDir)
            else:
                filelist.append(newDir)
        else:
            if recursive:
                listxfile(listdir=newDir, exts=exts, recursive=True, filelist=filelist)

    return filelist


def pathjoin(*kwargs):
    """Joint strings to a path.

    Parameters
    ----------
    *kwargs
        strings.

    Returns
    -------
    str
        The joined path string.
    """

    filesep = os.path.sep
    filepath = ''
    for k in kwargs:
        filepath += filesep + k
    return filepath[1:]


def fileparts(file):
    r"""Filename parts

    Returns the path, file name, and file name extension for the specified :attr:`file`.
    The :attr:`file` input is the name of a file or folder, and can include a path and
    file name extension.

    Parameters
    ----------
    file : str
        The name string of a file or folder.

    Returns
    -------
    filepath : str
        The path string of the file or folder.
    name : str
        The name string of the file or folder.
    ext : str
        The extension string of the file.

    """

    filepath, filename = os.path.split(file)
    name, ext = os.path.splitext(filename)
    return filepath, name, ext


def readtxt(filepath, mode=None):
    """Read a text file.

    Parameters
    ----------
    filepath : str
        The path string of the file.
    mode : str or None, optional
        ``'line'`` --> reading line-by-line.
        ``'all'`` --> reading all content. (default)

    Returns
    -------
    str
        File content.
    """
    if mode is None:
        mode = 'all'

    with open(filepath, "r") as f:
        if mode == 'all':
            data = f.read()
        if mode == 'line':
            data = f.readlines()
            N = len(data)
            for n in range(N):
                data[n] = data[n].strip('\n')
        f.close()
    return data


def readnum(filepath, pmain='Train', psub='loss: ', vfn=float, nshots=None):
    """Read a file and extract numbers in it.

    Parameters
    ----------
    filepath : str
        The path string of the file.
    pmain : str, optional
        The matching pattern string, such as '--->Train'.
    psub : str, optional
        The sub-matching pattern string, such as 'loss'.
    vfn : function, optional
        The function for formating the numbers. ``float`` --> convert to float number; ``int`` --> convert to integer number..., The default is ``float``.
    nshots : None, optional
        The number of shots of sub-matching pattern.

    Returns
    -------
    list
        The list of numbers.
    """
    if nshots is None:
        nshots = float('Inf')
    v = []
    cnt = 1
    with open(filepath, 'r') as f:
        while True:
            datastr = f.readline()
            posmain = datastr.find(pmain)
            if datastr == '':
                break
            if posmain > -1:
                aa = re.findall(psub + r'-?\d+\.?\d*e*E?[-+]?\d*', datastr)
                if aa != []:
                    v.append(vfn(aa[0][len(psub):]))
                    cnt += 1
            if cnt > nshots:
                break
    return v


def readcsv(filepath, sep=None, vfn=None, nlines=None):
    """Read a csv file and extract numbers in it.

    Parameters
    ----------
    filepath : str
        The path string of the file.
    sep : str, optional
        The separation character. Such as ``','`` or ``' '``. If None (default) or ``''`` (empty) return a list of all the lines.
    vfn : function or None, optional
        The function for formating the numbers. ``float`` --> convert to float number; ``int`` --> convert to integer number..., The default is :obj:`None`, which means won't converted, string format.
    nlines : None, optional
        The number of lines for reading, the default is None, which means all the lines.

    Returns
    -------
    list
        The list of numbers or strings.
    """
    if nlines is None:
        nlines = float('Inf')
    numbers = []
    cnt = 1
    with open(filepath, 'r') as f:
        while True:
            datastr = f.readline().strip()
            if not datastr:
                break
            if datastr != '':
                if (sep is not None) and sep != '':
                    datalist = datastr.split(sep)
                    if vfn is None:
                        numbers.append(datalist)
                    else:
                        numbers.append([vfn(v) for v in datalist])
                else:
                    numbers.append(datastr)
                cnt += 1
            if cnt > nlines:
                break
    return numbers


def readsec(filepath, pmain='Train', psub='time: ', vfn=int, nshots=None):
    """Read a file and extract seconds in it.

        ``hh:mm:ss``  -->  ``hh*3600 + mm*60 + ss``

    Parameters
    ----------
    filepath : str
        The path string of the file.
    pmain : str, optional
        The matching pattern string, such as '--->Train'.
    psub : str, optional
        The sub-matching pattern string, such as 'loss'.
    vfn : function or None, optional
        The function for formating the numbers. ``float`` --> convert to float number; ``int`` --> convert to integer number.
    nshots : None, optional
        The number of shots of sub-matching pattern.

    Returns
    -------
    list
        The list of seconds.
    """
    if nshots is None:
        nshots = float('Inf')
    v = []
    cnt = 1
    with open(filepath, 'r') as f:
        while True:
            datastr = f.readline()
            posmain = datastr.find(pmain)
            if datastr == '':
                break
            if posmain > -1:
                aa = re.findall(psub + r'\d+:\d+:\d+', datastr)
                if aa != []:
                    v.append(vfn(str2sec(aa[0][len(psub):])))
                    cnt += 1
            if cnt > nshots:
                break
    return v


if __name__ == '__main__':

    files = listxfile(listdir='/home/liu/', exts=None, recursive=False, filelist=[])
    print(files)

    filepath = pathjoin('a', 'b', 'c', '.d')
    print(filepath)

    filepath = 'data/files/log/log.log'

    v = readnum(filepath, pmain='Train', psub='loss: ', vfn=float, nshots=10)
    print(v)

    print('===================')
    datastr = ' 1  1.3e-02  3.4e+03fd7.6lo6.'
    print(re.findall(r'-?\d+\.?\d*e*E?[-+]?\d*', datastr))


from __future__ import division

from collections import defaultdict
import itertools
import numpy as np
import six

# https://github.com/chainer/chainercv/blob/master/chainercv/evaluations/eval_detection_voc.py

def bbox_iou(bbox_a, bbox_b):
    r"""Calculate the Intersection of Unions (IoUs) between bounding boxes.
    IoU is calculated as a ratio of area of the intersection
    and area of the union.
    This function accepts both :obj:`numpy.ndarray` and :obj:`cupy.ndarray` as
    inputs. Please note that both :obj:`bbox_a` and :obj:`bbox_b` need to be
    same type.
    The output is same type as the type of the inputs.
    Args:
        bbox_a (array): An array whose shape is :math:`(N, 4)`.
            :math:`N` is the number of bounding boxes.
            The dtype should be :obj:`numpy.float32`.
        bbox_b (array): An array similar to :obj:`bbox_a`,
            whose shape is :math:`(K, 4)`.
            The dtype should be :obj:`numpy.float32`.
    Returns:
        array:
        An array whose shape is :math:`(N, K)`. \
        An element at index :math:`(n, k)` contains IoUs between \
        :math:`n` th bounding box in :obj:`bbox_a` and :math:`k` th bounding \
        box in :obj:`bbox_b`.
    """
    if bbox_a.shape[1] != 4 or bbox_b.shape[1] != 4:
        raise IndexError

    # top left
    tl = np.maximum(bbox_a[:, None, :2], bbox_b[:, :2])
    # bottom right
    br = np.minimum(bbox_a[:, None, 2:], bbox_b[:, 2:])

    area_i = np.prod(br - tl, axis=2) * (tl < br).all(axis=2)
    area_a = np.prod(bbox_a[:, 2:] - bbox_a[:, :2], axis=1)
    area_b = np.prod(bbox_b[:, 2:] - bbox_b[:, :2], axis=1)
    return area_i / (area_a[:, None] + area_b - area_i)


def eval_detection_voc(
        pred_bboxes,
        pred_labels,
        pred_scores,
        gt_bboxes,
        gt_labels,
        gt_difficults=None,
        iou_thresh=0.5,
        use_07_metric=False):
    r"""Calculate average precisions based on evaluation code of PASCAL VOC.

    This function evaluates predicted bounding boxes obtained from a dataset
    which has :math:`N` images by using average precision for each class.
    The code is based on the evaluation code used in PASCAL VOC Challenge.

    Args:
        pred_bboxes (iterable of numpy.ndarray): An iterable of :math:`N`
            sets of bounding boxes.
            Its index corresponds to an index for the base dataset.
            Each element of :obj:`pred_bboxes` is a set of coordinates
            of bounding boxes. This is an array whose shape is :math:`(R, 4)`,
            where :math:`R` corresponds
            to the number of bounding boxes, which may vary among boxes.
            The second axis corresponds to
            :math:`y_{min}, x_{min}, y_{max}, x_{max}` of a bounding box.
        pred_labels (iterable of numpy.ndarray): An iterable of labels.
            Similar to :obj:`pred_bboxes`, its index corresponds to an
            index for the base dataset. Its length is :math:`N`.
        pred_scores (iterable of numpy.ndarray): An iterable of confidence
            scores for predicted bounding boxes. Similar to :obj:`pred_bboxes`,
            its index corresponds to an index for the base dataset.
            Its length is :math:`N`.
        gt_bboxes (iterable of numpy.ndarray): An iterable of ground truth
            bounding boxes
            whose length is :math:`N`. An element of :obj:`gt_bboxes` is a
            bounding box whose shape is :math:`(R, 4)`. Note that the number of
            bounding boxes in each image does not need to be same as the number
            of corresponding predicted boxes.
        gt_labels (iterable of numpy.ndarray): An iterable of ground truth
            labels which are organized similarly to :obj:`gt_bboxes`.
        gt_difficults (iterable of numpy.ndarray): An iterable of boolean
            arrays which is organized similarly to :obj:`gt_bboxes`.
            This tells whether the
            corresponding ground truth bounding box is difficult or not.
            By default, this is :obj:`None`. In that case, this function
            considers all bounding boxes to be not difficult.
        iou_thresh (float): A prediction is correct if its Intersection over
            Union with the ground truth is above this value.
        use_07_metric (bool): Whether to use PASCAL VOC 2007 evaluation metric
            for calculating average precision. The default value is
            :obj:`False`.

    Returns:
        dict:

        The keys, value-types and the description of the values are listed
        below.

        - **ap** (ndarray): An array of average precisions. \
            The :math:`l`-th value corresponds to the average precision \
            for class :math:`l`. If class :math:`l` does not exist in \
            either :obj:`pred_labels` or :obj:`gt_labels`, the corresponding \
            value is set to :obj:`numpy.nan`.
        - **map** (float): The average of Average Precisions over classes.

    """

    prec, rec = calc_detection_voc_prec_rec(pred_bboxes,
                                            pred_labels,
                                            pred_scores,
                                            gt_bboxes,
                                            gt_labels,
                                            gt_difficults,
                                            iou_thresh=iou_thresh)

    ap = calc_detection_voc_ap(prec, rec, use_07_metric=use_07_metric)

    return {'ap': ap, 'map': np.nanmean(ap)}


def calc_detection_voc_prec_rec(
        pred_bboxes, pred_labels, pred_scores, gt_bboxes, gt_labels,
        gt_difficults=None,
        iou_thresh=0.5):
    r"""Calculate precision and recall based on evaluation code of PASCAL VOC.

    This function calculates precision and recall of
    predicted bounding boxes obtained from a dataset which has :math:`N`
    images.
    The code is based on the evaluation code used in PASCAL VOC Challenge.

    Args:
        pred_bboxes (iterable of numpy.ndarray): An iterable of :math:`N`
            sets of bounding boxes.
            Its index corresponds to an index for the base dataset.
            Each element of :obj:`pred_bboxes` is a set of coordinates
            of bounding boxes. This is an array whose shape is :math:`(R, 4)`,
            where :math:`R` corresponds
            to the number of bounding boxes, which may vary among boxes.
            The second axis corresponds to
            :math:`y_{min}, x_{min}, y_{max}, x_{max}` of a bounding box.
        pred_labels (iterable of numpy.ndarray): An iterable of labels.
            Similar to :obj:`pred_bboxes`, its index corresponds to an
            index for the base dataset. Its length is :math:`N`.
        pred_scores (iterable of numpy.ndarray): An iterable of confidence
            scores for predicted bounding boxes. Similar to :obj:`pred_bboxes`,
            its index corresponds to an index for the base dataset.
            Its length is :math:`N`.
        gt_bboxes (iterable of numpy.ndarray): An iterable of ground truth
            bounding boxes
            whose length is :math:`N`. An element of :obj:`gt_bboxes` is a
            bounding box whose shape is :math:`(R, 4)`. Note that the number of
            bounding boxes in each image does not need to be same as the number
            of corresponding predicted boxes.
        gt_labels (iterable of numpy.ndarray): An iterable of ground truth
            labels which are organized similarly to :obj:`gt_bboxes`.
        gt_difficults (iterable of numpy.ndarray): An iterable of boolean
            arrays which is organized similarly to :obj:`gt_bboxes`.
            This tells whether the
            corresponding ground truth bounding box is difficult or not.
            By default, this is :obj:`None`. In that case, this function
            considers all bounding boxes to be not difficult.
        iou_thresh (float): A prediction is correct if its Intersection over
            Union with the ground truth is above this value..

    Returns:
        tuple of two lists:
        This function returns two lists: :obj:`prec` and :obj:`rec`.

        * :obj:`prec`: A list of arrays. :obj:`prec[l]` is precision \
            for class :math:`l`. If class :math:`l` does not exist in \
            either :obj:`pred_labels` or :obj:`gt_labels`, :obj:`prec[l]` is \
            set to :obj:`None`.
        * :obj:`rec`: A list of arrays. :obj:`rec[l]` is recall \
            for class :math:`l`. If class :math:`l` that is not marked as \
            difficult does not exist in \
            :obj:`gt_labels`, :obj:`rec[l]` is \
            set to :obj:`None`.

    """

    pred_bboxes = iter(pred_bboxes)
    pred_labels = iter(pred_labels)
    pred_scores = iter(pred_scores)
    gt_bboxes = iter(gt_bboxes)
    gt_labels = iter(gt_labels)
    if gt_difficults is None:
        gt_difficults = itertools.repeat(None)
    else:
        gt_difficults = iter(gt_difficults)

    n_pos = defaultdict(int)
    score = defaultdict(list)
    match = defaultdict(list)

    for pred_bbox, pred_label, pred_score, gt_bbox, gt_label, gt_difficult in \
            six.moves.zip(
                pred_bboxes, pred_labels, pred_scores,
                gt_bboxes, gt_labels, gt_difficults):

        if gt_difficult is None:
            gt_difficult = np.zeros(gt_bbox.shape[0], dtype=bool)

        for l in np.unique(np.concatenate((pred_label, gt_label)).astype(int)):
            pred_mask_l = pred_label == l
            pred_bbox_l = pred_bbox[pred_mask_l]
            pred_score_l = pred_score[pred_mask_l]
            # sort by score
            order = pred_score_l.argsort()[::-1]
            pred_bbox_l = pred_bbox_l[order]
            pred_score_l = pred_score_l[order]

            gt_mask_l = gt_label == l
            gt_bbox_l = gt_bbox[gt_mask_l]
            gt_difficult_l = gt_difficult[gt_mask_l]

            n_pos[l] += np.logical_not(gt_difficult_l).sum()
            score[l].extend(pred_score_l)

            if len(pred_bbox_l) == 0:
                continue
            if len(gt_bbox_l) == 0:
                match[l].extend((0,) * pred_bbox_l.shape[0])
                continue

            # VOC evaluation follows integer typed bounding boxes.
            pred_bbox_l = pred_bbox_l.copy()
            pred_bbox_l[:, 2:] += 1
            gt_bbox_l = gt_bbox_l.copy()
            gt_bbox_l[:, 2:] += 1

            iou = bbox_iou(pred_bbox_l, gt_bbox_l)
            gt_index = iou.argmax(axis=1)
            # set -1 if there is no matching ground truth
            gt_index[iou.max(axis=1) < iou_thresh] = -1
            del iou

            selec = np.zeros(gt_bbox_l.shape[0], dtype=bool)
            for gt_idx in gt_index:
                if gt_idx >= 0:
                    if gt_difficult_l[gt_idx]:
                        match[l].append(-1)
                    else:
                        if not selec[gt_idx]:
                            match[l].append(1)
                        else:
                            match[l].append(0)
                    selec[gt_idx] = True
                else:
                    match[l].append(0)

    for iter_ in (
            pred_bboxes, pred_labels, pred_scores,
            gt_bboxes, gt_labels, gt_difficults):
        if next(iter_, None) is not None:
            raise ValueError('Length of input iterables need to be same.')

    n_fg_class = max(n_pos.keys()) + 1
    prec = [None] * n_fg_class
    rec = [None] * n_fg_class

    for l in n_pos.keys():
        score_l = np.array(score[l])
        match_l = np.array(match[l], dtype=np.int8)

        order = score_l.argsort()[::-1]
        match_l = match_l[order]

        tp = np.cumsum(match_l == 1)
        fp = np.cumsum(match_l == 0)

        # If an element of fp + tp is 0,
        # the corresponding element of prec[l] is nan.
        prec[l] = tp / (fp + tp)
        # If n_pos[l] is 0, rec[l] is None.
        if n_pos[l] > 0:
            rec[l] = tp / n_pos[l]

    return prec, rec


def calc_detection_voc_ap(prec, rec, use_07_metric=False):
    r"""Calculate average precisions based on evaluation code of PASCAL VOC.

    This function calculates average precisions
    from given precisions and recalls.
    The code is based on the evaluation code used in PASCAL VOC Challenge.

    Args:
        prec (list of numpy.array): A list of arrays.
            :obj:`prec[l]` indicates precision for class :math:`l`.
            If :obj:`prec[l]` is :obj:`None`, this function returns
            :obj:`numpy.nan` for class :math:`l`.
        rec (list of numpy.array): A list of arrays.
            :obj:`rec[l]` indicates recall for class :math:`l`.
            If :obj:`rec[l]` is :obj:`None`, this function returns
            :obj:`numpy.nan` for class :math:`l`.
        use_07_metric (bool): Whether to use PASCAL VOC 2007 evaluation metric
            for calculating average precision. The default value is
            :obj:`False`.

    Returns:
        ~numpy.ndarray:
        This function returns an array of average precisions.
        The :math:`l`-th value corresponds to the average precision
        for class :math:`l`. If :obj:`prec[l]` or :obj:`rec[l]` is
        :obj:`None`, the corresponding value is set to :obj:`numpy.nan`.

    """

    n_fg_class = len(prec)
    ap = np.empty(n_fg_class)
    for l in six.moves.range(n_fg_class):
        if prec[l] is None or rec[l] is None:
            ap[l] = np.nan
            continue

        if use_07_metric:
            # 11 point metric
            ap[l] = 0
            for t in np.arange(0., 1.1, 0.1):
                if np.sum(rec[l] >= t) == 0:
                    p = 0
                else:
                    p = np.max(np.nan_to_num(prec[l])[rec[l] >= t])
                ap[l] += p / 11.
        else:
            # correct AP calculation
            # first append sentinel values at the end
            mpre = np.concatenate(([0], np.nan_to_num(prec[l]), [0]))
            mrec = np.concatenate(([0], rec[l], [1]))
            mpre = np.maximum.accumulate(mpre[::-1])[::-1]

            # to calculate area under PR curve, look for points
            # where X axis (recall) changes value
            i = np.where(mrec[1:] != mrec[:-1])[0]

            # and sum (\Delta recall) * prec
            ap[l] = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap
